﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: ToneRenderer.frag 537 2012-04-04 15:35:42Z cr333 $
*/
#version 120
#extension GL_EXT_gpu_shader4 : enable

uniform sampler2D colour;
uniform sampler2D depth;

// direction or position of light
uniform vec3 lightDirection;

// is lightDirection actually an absolute position?
uniform bool isPosition;

// Phong shading parameters
uniform float ambient;
uniform float diffuse;
uniform float specular;
uniform float shininess;

uniform vec4 toonStep;

// output selector
uniform int toneOutput;

// forward declaration of mesh vertex shader functions (from Mesh.vert)
vec3 getVertex(sampler2D depthmap, vec2 position);
vec3 getVertexNormal(sampler2D depthmap, vec2 position, vec3 coords);

void main (void)
{
	// vertex position and normal, in world coordinates
	vec2 pos = gl_FragCoord.xy - 1; // minus one to distribute border evenly on all sides
	vec3 vertex = getVertex(depth, pos);
	vec3 normal = getVertexNormal(depth, pos, vertex);
		
	// compute normal and vectors to light, viewer and reflected ray (all normalised)
	vec3 N = normalize(normal);
	vec3 L = normalize(lightDirection - (isPosition ? -vertex : vec3(0.0))); // translate if using absolute world coordinates
	vec3 V = vec3(0.0, 0.0, -1.0);
	vec3 R = normalize(-reflect(L, N));

	// diffuse tone image (ambient + diffuse terms)
	float diffuseTone = ambient + diffuse * max(dot(N, L), 0.0);
	
	// specular tone image (specular + shininess)
	float specularTone = pow(max(dot(R, V), 0.0), shininess);
	
	// classic Phong shading (for cel shading)
	float phong = (diffuseTone + specularTone) / (ambient + diffuse + specular);
	
	// hard-coded toon shading
	//float toon = 0.4 + 0.4 * smoothstep(0.2, 0.4, phong);
	//float toon = 0.7 + 0.3 * smoothstep(0.4, 0.6, phong);
	float toon = toonStep.x + toonStep.y * smoothstep(toonStep.z, toonStep.w, phong);
	
	switch(toneOutput)
	{
		case 0: gl_FragColor = vec4(diffuseTone, specularTone, toon, 1); break; // create multi-channel tone image
	
		// individual layers
		case 1: gl_FragColor = vec4(vec3(diffuseTone), 1); break;
		case 2: gl_FragColor = vec4(vec3(specularTone), 1); break;
		case 3: gl_FragColor = vec4(vec3(toon), 1); break;

		// additional debug outputs
		case 4: gl_FragColor = vec4(N, 1); break; // plain normals
		case 5: gl_FragColor = vec4(0.5 * N + vec3(0.5), 1); break; // normals mapped to visible colours
		
		case 6: // alignment of depth (Sobel-filtered) and colour
			vec3 col = texture2D(colour, pos / textureSize2D(colour, 0)).rgb;
			float zzz = 4 * texture2D(depth, (pos + vec2( 0,  0)) / textureSize2D(depth, 0)).z -
							texture2D(depth, (pos + vec2(-1,  0)) / textureSize2D(depth, 0)).z -
							texture2D(depth, (pos + vec2( 0, -1)) / textureSize2D(depth, 0)).z -
							texture2D(depth, (pos + vec2( 0,  1)) / textureSize2D(depth, 0)).z -
							texture2D(depth, (pos + vec2( 1,  0)) / textureSize2D(depth, 0)).z;
			gl_FragColor = vec4(vec3(clamp(10 * zzz + 1, 0, 2)) * col, 1);
			break;

		case 7: // half-colour overlay
			col = texture2D(colour, pos / textureSize2D(colour, 0)).rgb;
			gl_FragColor = vec4(mix(vec3(diffuseTone), col, 0.5), 1);
			break;
			
		case 8: // basic 'toon relighting'
			col = texture2D(colour, pos / textureSize2D(colour, 0)).rgb;
			gl_FragColor = vec4(toon * col, 1);
			break;
			
		case 9: // just colour image
			col = texture2D(colour, pos / textureSize2D(colour, 0)).rgb;
			gl_FragColor = vec4(col, 1);
			break;
			
		case 10: // just depth map
			vec3 depth_ = texture2D(depth, pos / textureSize2D(depth, 0)).rgb;
			gl_FragColor = vec4(depth_.zzz, 1);
			break;
			
		case 11: // raw point cloud
			gl_FragColor = texture2D(depth, pos / textureSize2D(depth, 0));
			break;
			
		case 12: // just distance map
			float dist = texture2D(depth, pos / textureSize2D(depth, 0)).a;
			gl_FragColor = vec4(vec3(0.5 * dist), 1);
			break;
	}
	
	//// render profile of a pixel column
	//vec3 d = texture2D(depth, vec2(400, gl_FragCoord.y) / textureSize2D(depth, 0)).zzz;
	//d = 100 * abs(d - gl_FragCoord.x / textureSize2D(depth, 0).x);
	//gl_FragColor = vec4(d, 1);
}