﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: ToneBlurNode.cs 537 2012-04-04 15:35:42Z cr333 $
*/
using System.Collections.Generic;
using System.ComponentModel;
using OpenTK.Graphics.OpenGL;
using VideoLib.OpenGL;

namespace RGBZcamera.Library.AbstractedLines
{
	public class ToneBlurNode : FboNode
	{
		[Uniform(), Category("Gaussian blur"), DefaultValue(10.0f), Description("Standard deviation of the Gaussian blur.")]
		public float Sigma { get; set; }

		public ToneBlurNode()
		{
			Sigma = 10;

			AddDefaultPins();
		}

		#region Rendering functions
//=================================================================================================

		private Texture intermediateTexture;

		/// <summary>Initialises the OpenGL state.</summary>
		public override void Init()
		{
			// initialise intermediate texture
			intermediateTexture = new Texture(128, 128) { Wrap = TextureWrapMode.ClampToEdge };
			
			// load, compile and link shaders
			programID = Utils.LinkShaders(
				Utils.CreateShaderFromResource("AbstractedLines.Shaders.ToneBlur.frag")
			);

			// map textures
			textureMap = new List<TextureBinding> {
				new TextureBinding { Pin = null, Uniform = "input", Unit = 0 },
			};

			base.Init();
		}

		public override TextureFrame Draw(Dictionary<string, TextureFrame> inputs)
		{
			return Draw(inputs["in"]);
		}
			
		private TextureFrame Draw(TextureFrame texture)
		{
			// resize intermediate texture if necessary (note: transposed relative to input texture)
			intermediateTexture.ReAlloc(texture.Height, texture.Width);

			// create output texture to fit input texture
			var resultTexture = TextureFrame.CreateToMatch(texture);
			resultTexture.Wrap = TextureWrapMode.ClampToEdge;
			resultTexture.Alloc();

			UpdateUniforms();

			// first pass of Gaussian blur
			texture.Bind(TextureUnit.Texture0);
			DrawScreenQuad(intermediateTexture);

			// second pass of Gaussian blur
			intermediateTexture.Bind(TextureUnit.Texture0);
			DrawScreenQuad(resultTexture);

			// cleanup
			intermediateTexture.Unbind();

			return resultTexture;
		}

//=================================================================================================
		#endregion // Rendering functions
	}
}
