﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: DistanceImageToPointCloud.cs 537 2012-04-04 15:35:42Z cr333 $
*/
using System.Collections.Generic;
using VideoLib.OpenGL;
using VideoLib.Vision;
using OpenTK;
using System.ComponentModel;

namespace RGBZcamera.Library
{
	/// <summary>Converts a depth or distance map into an XYZD point cloud.</summary>
	public class DistanceImageToPointCloud : FboNode
	{
		/// <summary>Describes the meaning of values as either distances or depths.</summary>
		public enum DistanceImageType
		{
			/// <summary>A distance map stores the distance along the viewing ray in each pixel (also known as 'range').</summary>
			DistanceMap = 0,

			/// <summary>A depth map stores the distance along the camera's principal axis (also known as orthogonal distance, or 'z').</summary>
			DepthMap = 1,
		}

		/// <summary>Gets or sets the intrinsics used to project a depth/distance map to 3D space.</summary>
		[Description("Intrinsic parameters used to project a depth or distance map to 3D space.")]
		public IntrinsicParameters Intrinsics { get; set; }

		#region Private properties to pass intrinsic parameters through to the shader.

		[Uniform()]
		private Vector2 fc { get { return new Vector2(Intrinsics.Fx, Intrinsics.Fy); } }

		[Uniform()]
		private Vector2 cc { get { return new Vector2(Intrinsics.Cx, Intrinsics.Cy); } }

		#endregion

		/// <summary>Gets or sets the scaling factor to apply.</summary>
		[Uniform(), Description("Depth/distance scale factor.")]
		public float DistanceScale { get; set; }

		/// <summary>Gets or sets the type of the image as either a depth or distance map.</summary>
		[Uniform(), Description("The type of image: either a depth or distance map.")]
		public DistanceImageType MapType { get; set; }

		/// <summary>Initialises a new instance of the <see cref="DistanceImageToPointCloud"/> class.</summary>
		public DistanceImageToPointCloud()
		{
			// default settings
			DistanceScale  = 1;
			MapType        = DistanceImageType.DistanceMap; // default for SR4000; Kinect/OpenNI needs DepthMap

			AddDefaultPins();
		}

		/// <summary>Initialises the OpenGL state.</summary>
		public override void Init()
		{
			// load, compile and link shaders
			programID = Utils.LinkShaders(
				Utils.CreateShaderFromResource("Shaders.DistanceImageToPointCloud.frag")
			);

			// map textures
			textureMap = new List<TextureBinding> {
				new TextureBinding { Pin = "in", Uniform = "input", Unit = 0 },
			};

			base.Init();
		}

		/// <summary>Draws to the framebuffer object using the given input textures.</summary>
		/// <param name="inputs">The input textures.</param>
		/// <returns>The resulting texture.</returns>
		public override TextureFrame Draw(Dictionary<string, TextureFrame> inputs)
		{
			AssertFrameProperties(inputs, FrameProps.Size);

			// allocate a texture that is identical to the input
			var resultTexture = new TextureFrame(inputs["in"].Width, inputs["in"].Height)
			{
				InternalFormat = OpenTK.Graphics.OpenGL.PixelInternalFormat.Rgba32f,
				Filter         = OpenTK.Graphics.OpenGL.TextureMagFilter.Nearest, // important for downstream processing (ReprojectionNode)
				SequenceNumber = inputs["in"].SequenceNumber,
			};
			resultTexture.Alloc();

			Draw(inputs, resultTexture.Texture);

			return resultTexture;
		}
	}
}
