﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: DepthBilateral.frag 537 2012-04-04 15:35:42Z cr333 $
*/
#extension GL_EXT_gpu_shader4 : enable

// input texture
uniform sampler2D input;

// parameters
uniform float spatialSigma;
uniform float rangeSigma;

void main()
{
	// look up texture size once and reuse it later,
	// assuming every texture has the same size
	vec2 size = vec2(textureSize2D(input, 0));

	// precompute 1 / sigma^2
	float iSigmaS = 1.0 / (2.0 * spatialSigma * spatialSigma);
	float iSigmaR = 1.0 / (2.0 * rangeSigma   * rangeSigma);

	// truncate Gaussian after 2 sigma
	int radius = int(ceil(2.0 * spatialSigma));

	// colour value at the central pixel
	vec3 centre = texture2D(input, gl_FragCoord.xy / size).rgb;
	
	// homogeneous accumulator of values, i.e. (wr, wg, wb, w)
	vec4 acc = vec4(0);

	for(int dy = -radius; dy <= radius; dy++)
	{
		for(int dx = -radius; dx <= radius; dx++)
		{
			vec2 coords = gl_FragCoord.xy + vec2(dx, dy);

			// colour value at the current pixel
			vec3 current = texture2D(input, coords / size).rgb;

			// compute pixel weight, using exp(a) * exp(b) = exp(a + b);
			float w = 0.0;

			// spatial weight
			vec2 deltaS = vec2(dx, dy);
			w += iSigmaS * dot(deltaS, deltaS);
			
			// range weight
			vec3 deltaR = current - centre;
			w += iSigmaR * dot(deltaR, deltaR);

			acc += exp(-w) * vec4(current, 1.0);
		}
	}

	// filtered colour (black if there were no valid data samples in the kernel)
	if(acc.w <= 0.0) gl_FragColor = vec4(0.0, 0.0, 0.0, 0.0);
	else             gl_FragColor = vec4(acc.rgb / acc.w, acc.z / acc.w);
}