﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: Relighter.frag 537 2012-04-04 15:35:42Z cr333 $
*/
#version 120
#extension GL_EXT_gpu_shader4 : enable

// input textures
uniform sampler2D colour;
uniform sampler2D points;

// remove illumination with these lighting coefficients
uniform float[9] lighting;

// scale albedo to achieve good light levels
uniform float scaleAlbedo;

// finally apply this environment map
uniform int envMap;

// rotate environment map
uniform vec3 rotation;

uniform bool outputMatte;
uniform vec4 backgroundPlane;
uniform float planeThreshold;

// forward declaration of mesh vertex shader functions (from Mesh.vert)
vec3 getVertex(sampler2D depthmap, vec2 position);
vec3 getVertexNormal(sampler2D depthmap, vec2 position, vec3 coords);

//// straightforward translation of Chenglei Wu's original code
//float[9] ComputeSphericalHarmonicsCoefficientsOriginal(vec3 normal)
//{
//	const float PI = 3.1415926;
//	const float[9] g_tmpkn = float[](0.8862, 1.0233, 0.4954, 0.0, -0.1108, 0.0, 0.0499, 0.0, -0.0285);
//	const float[9] g_scalefactor = float[](sqrt(1/PI)/2, -sqrt(3/PI/4), sqrt(3/PI/4), -sqrt(3/PI/4), sqrt(15/PI)/2, -sqrt(15/PI)/2, sqrt(5/PI)/4, -sqrt(15/PI)/2, sqrt(15/PI)/4);
//
//	float[9] vec_shval = float[](0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0);
//
//	float px = normal.x;
//	float py = normal.y;
//	float pz = normal.z;
//
//	float lenth  = sqrt(px * px + py * py + pz * pz);
//	float lenth2 = lenth  * lenth;
//	float lenth4 = lenth2 * lenth2;
//	float px2 = px * px;
//	float py2 = py * py;
//	float pz2 = pz * pz;
//
//	if(lenth <= 0.0)
//		return vec_shval;
//
//	float m_tmpshade = 0.0f;
//	vec_shval[0] = 1;
//	vec_shval[1] = py / lenth;
//	vec_shval[2] = pz / lenth;
//	vec_shval[3] = px / lenth;
//	vec_shval[4] = px * py / lenth2;
//	vec_shval[5] = py * pz /lenth2;
//	vec_shval[6] = (-px2-py2+2*pz2)/lenth2;
//	vec_shval[7] = pz * px / lenth2;
//	vec_shval[8] = (px2-py2)/lenth2;
//
//	for(int n = 0; n < 3; n++)
//	{
//		int m=0;
//		int m_tmpindtmp = 1+n*(n+1)+m-1;
//		vec_shval[m_tmpindtmp] =vec_shval[m_tmpindtmp]*g_tmpkn[n]*g_scalefactor[m_tmpindtmp];
//
//		for (m=1;m<(n+1);++m)
//		{
//			m_tmpindtmp = 1+n*(n+1)+m-1;
//			vec_shval[m_tmpindtmp] = vec_shval[m_tmpindtmp]*g_tmpkn[n]*g_scalefactor[m_tmpindtmp];
//
//			m_tmpindtmp = 1+n*(n+1)-m-1;
//			vec_shval[m_tmpindtmp] = vec_shval[m_tmpindtmp]*g_tmpkn[n]*g_scalefactor[m_tmpindtmp];
//		}
//	}
//
//	return vec_shval;
//}

// same as code above (within 0.001), but optimised and cleaned
float[9] ComputeSphericalHarmonicsCoefficients(vec3 normal)
{
	float len  = length(normal);
	float len2 = len * len;
	vec3  p2   = normal * normal;

	// ignore null normals
	if(len <= 0.0)
		return float[9](0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0);

	return float[9](
		 0.25,
		-0.5 * normal.y / len,
		 0.5 * normal.z / len,
		-0.5 * normal.x / len,
		 0.541265877 * normal.x * normal.y / len2,
		-0.541265877 * normal.y * normal.z / len2,
		 0.15625 * (-p2.x - p2.y + 2 * p2.z) / len2,
		-0.541265877 * normal.z * normal.x / len2,
		 0.270632939 * (p2.x - p2.y) / len2
	);
}

// coefficients from Ramamoorthi's 2001 SIGGRAPH paper

//const vec3[9] cathedral = vec3[](
	//vec3( .79,  .44,  .54),
	//vec3( .39,  .35,  .60),
	//vec3(-.34, -.18, -.27),
	//vec3(-.29, -.06,  .01),
	//vec3(-.11, -.05, -.12),
	//vec3(-.26, -.22, -.47),
	//vec3(-.16, -.09, -.15),
	//vec3( .56,  .21,  .14),
	//vec3( .21, -.05, -.30)
//);

//const vec3[9] grove = vec3[](
	//vec3( .38,  .43,  .45),
	//vec3( .29,  .36,  .41),
	//vec3( .04,  .03,  .01),
	//vec3(-.10, -.10, -.09),
	//vec3(-.06, -.06, -.04),
	//vec3( .01, -.01, -.05),
	//vec3(-.09, -.13, -.15),
	//vec3(-.06, -.05, -.04),
	//vec3( .02, -.00, -.05)
//);

//const vec3[9] basilica = vec3[](
	//vec3( .36,  .26,  .23),
	//vec3( .18,  .14,  .13),
	//vec3(-.02, -.01, -.00),
	//vec3( .03,  .02,  .01),
	//vec3( .02,  .01,  .00),
	//vec3(-.05, -.03, -.01),
	//vec3(-.09, -.08, -.07),
	//vec3( .01,  .00,  .00),
	//vec3(-.08, -.06,  .00)
//);

// coefficients computed by Chenglei Wu (different coordinate system; different scale for cathedral)

const vec3[9] basilica = vec3[9](
	vec3( 3.618,  2.618,  2.319),
	vec3(-0.348, -0.224, -0.102),
	vec3(-0.246, -0.100, -0.009),
	vec3( 1.755,  1.432,  1.255),
	vec3(-0.201, -0.145, -0.045),
	vec3(-0.045, -0.034,  0.004),
	vec3(-0.898, -0.759, -0.738),
	vec3(-0.468, -0.253, -0.116),
	vec3( 0.819,  0.320, -0.036)
);

const vec3[9] beach = vec3[9](
	vec3( 1.712,  1.734,  1.769),
	vec3( 1.129,  1.043,  0.857),
	vec3(-0.432, -0.429, -0.409),
	vec3( 0.791,  0.921,  1.100),
	vec3( 0.421,  0.425,  0.381),
	vec3(-0.407, -0.341, -0.249),
	vec3(-0.083, -0.056, -0.039),
	vec3(-0.210, -0.235, -0.257),
	vec3(-0.930, -0.780, -0.512)
);

const vec3[9] building = vec3[9](
	vec3( 3.179,  3.575,  3.789),
	vec3( 1.039,  1.175,  1.252),
	vec3(-0.055, -0.106, -0.126),
	vec3( 3.657,  4.126,  4.497),
	vec3( 1.291,  1.493,  1.688),
	vec3(-0.243, -0.280, -0.336),
	vec3(-0.705, -0.764, -0.868),
	vec3(-0.080, -0.135, -0.155),
	vec3( 2.863,  3.243,  3.595)
);

const vec3[9] campus = vec3[9](
	vec3( 1.962,  2.341,  2.449),
	vec3(-1.128, -0.929, -0.482),
	vec3(-0.257, -0.457, -0.686),
	vec3( 1.084,  1.386,  1.751),
	vec3(-0.444, -0.332, -0.107),
	vec3(-0.203, -0.171, -0.074),
	vec3(-0.971, -1.008, -0.910),
	vec3(-0.354, -0.546, -0.781),
	vec3(-1.668, -1.681, -1.316)
);

const vec3[9] cathedral = vec3[9](
	vec3(10.119,  5.586,  6.926),
	vec3( 3.751,  0.706, -0.131),
	vec3(-4.333, -2.315, -3.431),
	vec3( 5.038,  4.465,  7.748),
	vec3( 1.421,  0.650,  1.570),
	vec3(-7.127, -2.711, -1.742),
	vec3(-2.043, -1.235, -1.920),
	vec3(-3.331, -2.849, -6.058),
	vec3(-2.723,  0.703,  3.921)
);

const vec3[9] galileo = vec3[9](
	vec3( 5.169,  3.793,  3.528),
	vec3(-3.544, -2.700, -2.783),
	vec3(-1.136, -0.924, -0.862),
	vec3( 2.205,  1.701,  1.687),
	vec3(-3.191, -2.466, -2.557),
	vec3( 0.858,  0.679,  0.674),
	vec3(-1.848, -1.357, -1.416),
	vec3(-0.577, -0.470, -0.421),
	vec3(-2.765, -2.115, -2.085)
);

const vec3[9] grove = vec3[9](
	vec3( 3.776,  4.256,  4.503),
	vec3( 1.029,  1.027,  0.880),
	vec3( 0.391,  0.309,  0.112),
	vec3( 2.875,  3.575,  4.135),
	vec3( 0.615,  0.547,  0.389),
	vec3( 0.583,  0.514,  0.375),
	vec3(-0.930, -1.250, -1.524),
	vec3( 0.085, -0.140, -0.461),
	vec3(-0.211,  0.035,  0.443)
);

const vec3[9] kitchen = vec3[9](
	vec3( 1.595,  1.681,  1.817),
	vec3( 0.114,  0.086,  0.040),
	vec3( 1.048,  1.498,  1.937),
	vec3( 0.701,  0.780,  0.816),
	vec3( 0.244,  0.187,  0.120),
	vec3(-0.154, -0.028,  0.043),
	vec3( 0.953,  1.360,  1.793),
	vec3( 0.613,  0.970,  1.316),
	vec3( 0.078,  0.055,  0.076)
);

const vec3[9] uffizi = vec3[9](
	vec3( 3.165,  3.070,  3.495),
	vec3( 0.081,  0.084,  0.113),
	vec3(-0.027, -0.024, -0.031),
	vec3( 3.694,  3.666,  4.263),
	vec3( 0.186,  0.188,  0.246),
	vec3(-0.006, -0.004, -0.003),
	vec3(-2.781, -2.745, -3.177),
	vec3(-0.095, -0.088, -0.102),
	vec3( 2.411,  2.404,  2.837)
);


void main (void)
{
	// look up texture size once and reuse it later, assuming every texture has the same size
	vec2 size = vec2(textureSize2D(colour, 0));
	
	// fetch input pixel colour
	vec3 shaded = texture2D(colour, gl_FragCoord.xy / size).rgb;
	
	// compute normal
	vec2 pos = gl_FragCoord.xy - 1; // minus one to distribute border evenly on all sides
	vec3 vertex = getVertex(points, pos);
	vec3 normal = normalize(getVertexNormal(points, pos, vertex));
	
	// output alpha matte
	if(outputMatte)
	{
		// normalise plane normal (so that the distance to the plane is meaningful)
		vec4 plane = backgroundPlane / length(backgroundPlane.xyz);

		// return binary matte
		if(dot(vec4(vertex, 1.0), backgroundPlane) < planeThreshold) gl_FragColor = vec4(0.0, 0.0, 0.0, 1.0);
		else                                                         gl_FragColor = vec4(1.0, 1.0, 1.0, 1.0);
		return;
	}
	
//---- Step 1. Remove estimated illumination ------------------------------------------------------
	
	float[9] sph_coefficients = ComputeSphericalHarmonicsCoefficients(normal);
	
	float illumination = 0.0;
	for(int i = 0; i < 9; i++)
		illumination += lighting[i] * sph_coefficients[i];
		
	vec3 albedo  = shaded / illumination * scaleAlbedo;

//---- Step 2. Add new illumination ---------------------------------------------------------------

	// rotate normal to rotate environment map
	vec3 angle = radians(rotation);
	normal = mat3(cos(angle.y), 0.0, sin(angle.y), 0.0, 1.0, 0.0, -sin(angle.y), 0.0, cos(angle.y)) * normal; // rotate about y
	normal = mat3(1.0, 0.0, 0.0, 0.0, cos(angle.x), -sin(angle.x), 0.0, sin(angle.x), cos(angle.x)) * normal; // rotate about x
	normal = mat3(cos(angle.z), -sin(angle.z), 0.0, sin(angle.z), cos(angle.z), 0.0, 0.0, 0.0, 1.0) * normal; // rotate about z

	// Relight with spherical harmonics approximation of environment maps (Ramamoorthi & Hanrahan 2001)

	// Note that the environment maps use different coordinate systems to us, so
	// the normal needs to be transformed to produce correct-looking results:
	//
	//   * normal.xyz * vec3(-1, -1,  1) => matches base functions as shown in Figure 4 of the paper
	//   * normal.xzy * vec3( 1,  1, -1) => matches lightprobes in paper (using Ramamoorthi's coefficients)
	//   * normal.yzx * vec3(-1,  1, -1) => matches lightprobes in paper (using Chenglei's coefficients)
	sph_coefficients = ComputeSphericalHarmonicsCoefficients(normal.yzx * vec3(-1, 1, -1));

	vec3 newshading = vec3(0.0);
	switch(envMap)
	{
		case 0: newshading = vec3(0.4); break;
		case 1: for(int i = 0; i < 9; i++) newshading += 0.4  * basilica[i]  * sph_coefficients[i]; break;
		case 2: for(int i = 0; i < 9; i++) newshading += 0.6  * beach[i]     * sph_coefficients[i]; break;
		case 3: for(int i = 0; i < 9; i++) newshading += 0.2  * building[i]  * sph_coefficients[i]; break;
		case 4: for(int i = 0; i < 9; i++) newshading += 1.0  * campus[i]    * sph_coefficients[i]; break;
		case 5: for(int i = 0; i < 9; i++) newshading += 0.08 * cathedral[i] * sph_coefficients[i]; break;
		case 6: for(int i = 0; i < 9; i++) newshading += 0.8  * galileo[i]   * sph_coefficients[i]; break;
		case 7: for(int i = 0; i < 9; i++) newshading += 0.25 * grove[i]     * sph_coefficients[i]; break;
		case 8: for(int i = 0; i < 9; i++) newshading += 0.5  * kitchen[i]   * sph_coefficients[i]; break;
		case 9: for(int i = 0; i < 9; i++) newshading += 0.25 * uffizi[i]    * sph_coefficients[i]; break;
		case 10: newshading = vec3(max(0, dot(normal, normalize(vec3(0, 0, -1))))); break;
	}

	gl_FragColor = vec4(albedo * newshading, 1);
}
