﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: Reprojection.vert 537 2012-04-04 15:35:42Z cr333 $
*/
#version 120
#extension GL_EXT_gpu_shader4 : enable

// coordinates of vertices
uniform sampler2D coordinates;

//uniform float sobelThreshold;

// the vertex position, in viewing coordinates
varying vec3 v3;

varying vec3 n3;
varying float valid;
varying float sobel;
varying vec2 coords;

// forward declaration of mesh vertex shader functions (from Mesh.vert)
vec3 getVertex(sampler2D depthmap, vec2 position);
vec3 getVertexNormal(sampler2D depthmap, vec2 position, vec3 coords);


// Compute gradient magnitude of the depth image using two Sobel operators.
// Note: coordinates of the vertex are passed in to save a texture read.
float getSobel(vec3 vertexCoords)
{
	// 012
	// 345
	// 678
	float[9] z;

	// look up depth of the 8 neighbours
	vec2 texelSize = vec2(1) / textureSize2D(coordinates, 0);
	z[0] = texture2D(coordinates, (gl_Vertex.xy + vec2(-1.0, -1.0)) * texelSize).z;
	z[1] = texture2D(coordinates, (gl_Vertex.xy + vec2( 0.0, -1.0)) * texelSize).z;
	z[2] = texture2D(coordinates, (gl_Vertex.xy + vec2( 1.0, -1.0)) * texelSize).z;
	z[3] = texture2D(coordinates, (gl_Vertex.xy + vec2(-1.0,  0.0)) * texelSize).z;
	z[4] = vertexCoords.z;
	z[5] = texture2D(coordinates, (gl_Vertex.xy + vec2( 1.0,  0.0)) * texelSize).z;
	z[6] = texture2D(coordinates, (gl_Vertex.xy + vec2(-1.0,  1.0)) * texelSize).z;
	z[7] = texture2D(coordinates, (gl_Vertex.xy + vec2( 0.0,  1.0)) * texelSize).z;
	z[8] = texture2D(coordinates, (gl_Vertex.xy + vec2( 1.0,  1.0)) * texelSize).z;
	
	// compute gradient magnitude
	return sqrt(pow(z[0] + 2 * z[1] + z[2] - z[6] - 2 * z[7] - z[8], 2) + pow(z[0] + 2 * z[3] + z[6] - z[2] - 2 * z[5] - z[8], 2));
}

void main(void)  
{
	// vertex position and normal, in world coordinates
	vec3 vertex = getVertex(coordinates, gl_Vertex.xy);
	vec3 normal = getVertexNormal(coordinates, gl_Vertex.xy, vertex);

	// transform SR4000 coordinate system (x left, y up, z into the scene)
	// to calibrated camera coordinate system (x right, y down, z into the scene)
	vec4 v = vec4(-1, -1, 1, 1) * vec4(vertex, 1);

	// vertex in eye space
	vec4 v4 = gl_ModelViewMatrix * v;
	v3 = v4.xyz / v4.w;
	
	// normal vector at vertex, in view coordinates
	n3 = normalize(gl_NormalMatrix * normal);
	
	// gradient magnitude of the depth component
	sobel = getSobel(vertex);
	
	if(vertex.z > 4.5) valid = 0.0; // saturated pixels
	//else if(sobel > sobelThreshold) valid = 0.0; // edge pixels
	else               valid = 1.0; // all other pixels
	
	coords = gl_Vertex.xy + vec2(0.5); // vertices located at texels
	
	// fully-transformed coordinates
	gl_Position = gl_ProjectionMatrix * v4;
}