﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: XToonShader.cs 537 2012-04-04 15:35:42Z cr333 $
*/
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Reflection;
using OpenTK;
using OpenTK.Graphics.OpenGL;
using VideoLib.Frames;
using VideoLib.OpenGL;

namespace RGBZcamera.Library.XToon
{
	public enum AttributeMapType
	{
		DepthBased    = 1,
		Highlights    = 2,
		DebugDepthAM  = 3,
		DebugDiffuse  = 4,
		DebugSpecular = 5,
	}

	/// <summary>Predefined toon textures.</summary>
	public enum Look
	{
		Custom               =  0,
		DepthSmooth          =  1,
		DiscreteLOA          =  2,
		RecedingShadows      =  3,
		DepthBlur            =  4,
		IntermediateTones    =  5,
		GreenBrown           =  6,
		BlueShift            =  7,
		PinkHue              =  8,
		DecreasingSaturation =  9,
		SilhouetteSmoothing  = 10,
		FuzzyRendering       = 11,
		WhiteBacklight       = 12,
		PlasticHighlight     = 13,
		MetallicHighlight    = 14,
		RedHighlight         = 15,
	}

	/// <summary>Implements Barla et al.'s XToon shader.</summary>
	public class XToonShader : FboNode
	{
//=================================================================================================
		#region Properties

		/// <summary>Gets or sets the attribute map to use.</summary>
		[Uniform(), Category("X-Toon"), Description("Attribute map to use for detail control."), DisplayName("Attribute map")]
		public AttributeMapType AttributeMap { get; set; }

		/// <summary>Gets or sets the look from an existing toon texture or selects a custom one.</summary>
		[Category("X-Toon"), Description("Select from existing toon textures or a custom one.")]
		public Look Look
		{
			get { return look; }
			set { look = value; updateToonTexture = true; }
		}

		/// <summary>Gets or sets the toon texture.</summary>
		[ReadOnly(true), Category("X-Toon"), Description("Toon texture used for rendering."), DisplayName("Toon texture")]
		public Bitmap ToonTexture
		{
			get { return toonTexture; }
			set { look = Look.Custom; toonTexture = value; } // setting textureType also sets updateToonTexture=true
		}

		/// <summary>Gets or sets the depth range used in the depth-based attribute map.</summary>
		[Uniform(), Category("X-Toon"), DisplayName("Depth range"), Description("Input depth range for depth-based attribute map.")]
		public Vector2 DepthRange { get; set; }

		/// <summary>Gets or sets the light direction.</summary>
		[Uniform(), Category("Shading"), DisplayName("Light direction"), Description("Position or direction of the single point light source.")]
		public Vector3 LightDirection { get; set; }

		/// <summary>Gets or sets a value indicating whether the light source is a position and not a direction vector.</summary>
		[Uniform(), Category("Shading"), DisplayName("Is position?"), Description("If set to true, will treat light position as a direction instead.")]
		public bool IsPosition { get; set; }

		/// <summary>Gets or sets the diffuse reflectivity.</summary>
		[Uniform(), Category("Shading"), Description("The amount of diffuse illumination to use.")]
		public float Diffuse { get; set; }

		/// <summary>Gets or sets the specular reflectivity.</summary>
		[Uniform(), Category("Shading"), Description("The amount of specular illumination to use for the highlights attribute map.")]
		public float Specular { get; set; }

		/// <summary>Gets or sets Phong's shininess coefficient.</summary>
		[Uniform(), Category("Shading"), Description("The shininess of the material, the larger the more mirror-like.")]
		public float Shininess { get; set; }

		/// <summary>Gets or sets the offset to apply to the sequence number of frames.</summary>
		[Browsable(false)]
		public int SequenceNumberDelta { get; set; }


		[Uniform("toonTexture", Unit = 1)]
		private Texture ToonTex { get; set; }

		private bool updateToonTexture = true;
		private Look look;
		private Bitmap toonTexture;

		#endregion Properties
//=================================================================================================

		/// <summary>Initialises a new instance of the <see cref="XToonShader"/> class.</summary>
		public XToonShader()
		{
			// XToon settings
			AttributeMap = AttributeMapType.DepthBased;
			Look         = Look.DepthSmooth;
			DepthRange   = new Vector2(0.4f, 1.0f);

			// shading settings
			LightDirection = new Vector3(0.2f, 0.2f, -1.0f);
			IsPosition     = false;
			Diffuse        = 0.8f;
			Specular       = 1.0f;
			Shininess      = 20;

			AddDefaultPins();
		}

		/// <summary>Initialises the OpenGL state.</summary>
		public override void Init()
		{
			// load, compile and link shaders
			programID = Utils.LinkShaders(
				Utils.CreateShaderFromResource("Shaders.Mesh.vert", ShaderType.FragmentShader),
				Utils.CreateShaderFromResource("XToon.XToon.frag")
			);

			// map textures
			textureMap = new List<TextureBinding> {
				new TextureBinding { Pin = "in", Uniform = "points",      Unit = 0 },
				new TextureBinding { Pin = null, Uniform = "toonTexture", Unit = 1 },
			};

			base.Init();
		}

		/// <summary>Draws to the framebuffer object using the given input textures.</summary>
		/// <param name="inputs">The input textures.</param>
		/// <returns>The resulting texture.</returns>
		public override TextureFrame Draw(Dictionary<string, TextureFrame> inputs)
		{
			AssertFrameProperties(inputs, FrameProps.Size);

			// allocate a texture that is identical to the input
			var resultTexture = new TextureFrame(inputs["in"].Width, inputs["in"].Height) { SequenceNumber = inputs["in"].SequenceNumber + SequenceNumberDelta };
			resultTexture.Alloc();

			UpdateToonTexture();
			Draw(inputs, resultTexture.Texture);

			return resultTexture;
		}

		/// <summary>Updates the toon texture if necessary.</summary>
		private void UpdateToonTexture()
		{
			// skip texture update if not necessary
			if (ToonTex != null && updateToonTexture == false)
				return;

			// using predefined toon texture?
			if (Look != Look.Custom)
			{
				// try to read toon texture from resource
				var resourceName = String.Format("xtoon-{0:D2}.png", (int)Look);
				using (var stream = Assembly.GetExecutingAssembly().GetManifestResourceStream(GetType(), resourceName))
				{
					if (stream == null)
						throw new Exception(string.Format("Resource '{0}' not found.", resourceName));

					toonTexture = new Bitmap(stream);
				}
			}

			// create toon texture
			if (ToonTex == null)
				ToonTex = new Texture(ToonTexture.Width, ToonTexture.Height) { Wrap = TextureWrapMode.ClampToEdge };

			// update toon texture
			ToonTex.Create(new BitmapFrame(ToonTexture));
			updateToonTexture = false;
		}
	}
}
