﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: Settings.cs 537 2012-04-04 15:35:42Z cr333 $
*/
using System;
using System.ComponentModel;
using System.IO;
using VideoLib.FlyCapture;
using VideoLib.OpenNI;

namespace RGBZcamera.Recorder
{
	public enum ImageFormat
	{
		BMP,
		//GIF, // uses dithering
		PNG,
		RAW,
		TIF
	}

	public enum VideoSource
	{
		FleaSR4000,
		Kinect
	}

	public class Settings
	{
//---- Capture settings ---------------------------------------------------------------------------

		[Category("Capture Settings"), DefaultValue(false),
		Description("Write frames to disk asynchronously.")]
		public bool CaptureAsync { get; set; }

		[Category("Capture Settings"), DefaultValue(false),
		Description("Buffer frames in memory to reduce chances of de-sync.")]
		public bool CaptureBuffered { get; set; }

		[Category("Capture Settings"), DefaultValue(true),
		Description("Display the captured videos while recording.")]
		public bool CaptureDisplay { get; set; }

		[Category("Capture Settings"), DefaultValue(ImageFormat.RAW),
		Description("Frames will be captured and stored using this image format.")]
		public ImageFormat CaptureFormat { get; set; }

		[Category("Capture Settings"),
		Description("Path for storing captured data.")]
		public string CapturePath { get; set; }

		[Category("Capture Settings"),
		Description("Source(s) of the video streams.")]
		public VideoSource CaptureSource { get; set; }


//---- Flea settings ------------------------------------------------------------------------------

		[Category("Point Grey Research Flea2"), DefaultValue(1024), TypeConverter(typeof(ResolutionConverter)),
		Description("The video will be captured at this resolution.")]
		public int Width { get; set; }

		[Category("Point Grey Research Flea2"), DefaultValue(Formats.Yuv422),
		Description("The video will be captured in this format and converted later, if necessary.")]
		public Formats Format { get; set; }

		[Category("Point Grey Research Flea2"), DefaultValue(Framerates._15), TypeConverter(typeof(FramerateConverter)),
		Description("The video will be captured at this framerate.")]
		public Framerates Framerate { get; set; }

		[Category("Point Grey Research Flea2"), DefaultValue(false),
		Description("Enable auto exposure.")]
		public bool AutoExposureFlea { get; set; }

		[Category("Point Grey Research Flea2"), DefaultValue(0.0f),
		Description("Exposure setting in EV.")]
		public float Exposure { get; set; }

		[Category("Point Grey Research Flea2"), DefaultValue(false),
		Description("Enable auto gain.")]
		public bool AutoGain { get; set; }

		[Category("Point Grey Research Flea2"), DefaultValue(4.0f),
		Description("Gain setting in dB.")]
		public float Gain { get; set; }

		[Category("Point Grey Research Flea2"), DefaultValue(2.0f),
		Description("Gamma setting.")]
		public float Gamma { get; set; }

		[Category("Point Grey Research Flea2"), DefaultValue(false),
		Description("Enable auto saturation.")]
		public bool AutoSaturation { get; set; }

		[Category("Point Grey Research Flea2"), DefaultValue(120.0f),
		Description("Saturation setting in percent.")]
		public float Saturation { get; set; }

		[Category("Point Grey Research Flea2"), DefaultValue(false),
		Description("Enable auto shutter time.")]
		public bool AutoShutter { get; set; }

		[Category("Point Grey Research Flea2"), DefaultValue(30.0f),
		Description("Shutter time in ms.")]
		public float Shutter { get; set; }

		[Category("Point Grey Research Flea2"), DefaultValue(false),
		Description("Enable auto white balancing.")]
		public bool AutoWhiteBalance { get; set; }

		[Category("Point Grey Research Flea2"), DefaultValue(750),
		Description("Blue setting for manual white balance.")]
		public int WhiteBalanceBlue { get; set; }

		[Category("Point Grey Research Flea2"), DefaultValue(550),
		Description("Red setting for manual white balance.")]
		public int WhiteBalanceRed { get; set; }


//---- SwissRanger settings -----------------------------------------------------------------------

		[Category("SwissRanger 4000"), DefaultValue(false),
		Description("Auto exposure adaptively changes the integration time.")]
		public bool AutoExposureSR { get; set; }

		[Category("SwissRanger 4000"), DefaultValue(5.0f),
		Description("Sets the fixed integration time (if auto exposure is off).")]
		public float IntegrationTime { get; set; }

		[Category("SwissRanger 4000"), DefaultValue(true),
		Description("Record the amplitude image.")]
		public bool CaptureAmplitude { get; set; }

		[Category("SwissRanger 4000"), DefaultValue(false),
		Description("Record the confidence map.")]
		public bool CaptureConfidence { get; set; }

		[Category("SwissRanger 4000"), DefaultValue(false),
		Description("Record the point cloud.")]
		public bool CaptureCoords { get; set; }

		[Category("SwissRanger 4000"), DefaultValue(true),
		Description("Record the depth map.")]
		public bool CaptureDepth { get; set; }


//---- Kinect settings ----------------------------------------------------------------------------

		[Category("Microsoft Kinect"), DefaultValue(Streams.ImageAndDepth),
		Description("Which video streams to capture from the Kinect.")]
		public Streams Streams { get; set; }

		[Category("Microsoft Kinect"), DefaultValue(ImageModes.SXGA_15fps),
		Description("Resolution and frame rate of the Kinect's image stream.")]
		public ImageModes ImageMode { get; set; }

		[Category("Microsoft Kinect"), DefaultValue(IRModes.VGA_30fps),
		Description("Resolution and frame rate of the Kinect's infrared stream.")]
		public IRModes IRMode { get; set; }

		[Category("Microsoft Kinect"), DefaultValue(DepthModes.VGA_30fps),
		Description("Resolution and frame rate of the Kinect's depth stream.")]
		public DepthModes DepthMode { get; set; }

		[Category("Microsoft Kinect"), DefaultValue(SyncMode.DropFrames),
		Description("Synchronisation mode for multiple image streams.")]
		public SyncMode SyncMode { get; set; }

		public Settings()
		{
			// default capture settings
			CaptureAsync    = true;
			CaptureBuffered = false;
			CaptureDisplay  = true;
			CaptureFormat   = ImageFormat.RAW;
			CapturePath     = @"C:\data\captured\";
			CaptureSource   = VideoSource.Kinect;

			// Flea2 default settings
			Width            = 1024;
			Format           = Formats.Yuv422;
			Framerate        = Framerates._15;
			AutoExposureFlea = false;
			Exposure         = 0;
			AutoGain         = false;
			Gain             = 4;
			Gamma            = 2;
			AutoSaturation   = false;
			Saturation       = 120;
			AutoShutter      = false;
			Shutter          = 30;
			AutoWhiteBalance = false;
			WhiteBalanceBlue = 750;
			WhiteBalanceRed  = 550;

			// SwissRanger default settings
			AutoExposureSR    = false;
			IntegrationTime   = 5.0f;
			CaptureAmplitude  = true;
			CaptureConfidence = false;
			CaptureCoords     = false;
			CaptureDepth      = true;

			// Microsoft Kinect settings
			Streams   = Streams.ImageAndDepth;
			ImageMode = ImageModes.SXGA_15fps;
			IRMode    = IRModes.VGA_30fps;
			DepthMode = DepthModes.VGA_30fps;
			SyncMode  = SyncMode.DropFrames;
		}

		internal void SaveToFile(string filename)
		{
			using (TextWriter writer = File.CreateText(filename))
			{
				writer.WriteLine("Capture Settings");
				writer.WriteLine("----------------------------------------");
				writer.WriteLine("Date             : " + DateTime.Now.ToString("yyyy-MM-dd"));
				writer.WriteLine("Time             : " + DateTime.Now.ToString("HH:mm:ss"));
				writer.WriteLine("Asynchronous     : " + CaptureAsync);
				writer.WriteLine("Buffered         : " + CaptureBuffered);
				writer.WriteLine("Display          : " + CaptureDisplay);
				writer.WriteLine("Format           : " + CaptureFormat);
				writer.WriteLine("Path             : " + CapturePath);
				writer.WriteLine("Source           : " + CaptureSource);
				writer.WriteLine();

				if (CaptureSource == VideoSource.FleaSR4000)
				{
					writer.WriteLine("SwissRanger 4000");
					writer.WriteLine("----------------------------------------");
					writer.WriteLine("Auto exposure    : " + AutoExposureSR);
					if (!AutoExposureSR)
						writer.WriteLine("Integration time : " + IntegrationTime.ToString());
					writer.WriteLine("Capture ampl. map: " + CaptureAmplitude);
					writer.WriteLine("Capture conf. map: " + CaptureConfidence);
					writer.WriteLine("Capture coords   : " + CaptureCoords);
					writer.WriteLine("Capture depth map: " + CaptureDepth);
					writer.WriteLine();

					writer.WriteLine("PointGreyResearch Flea2");
					writer.WriteLine("----------------------------------------");
					writer.WriteLine("Resolution       : " + Width.ToString());
					writer.WriteLine("Format           : " + Format);
					writer.WriteLine("Framerate        : " + new FramerateConverter(typeof(Framerates)).ConvertToString(Framerate));
					writer.WriteLine("Auto Exposure    : " + AutoExposureFlea);
					if (!AutoExposureFlea)
						writer.WriteLine("Exposure         : " + Exposure.ToString());
					writer.WriteLine("Auto Gain        : " + AutoGain);
					if (!AutoGain)
						writer.WriteLine("Gain             : " + Gain.ToString());
					writer.WriteLine("Gamma            : " + Gamma.ToString());
					writer.WriteLine("Auto Saturation  : " + AutoSaturation);
					if (!AutoSaturation)
						writer.WriteLine("Saturation       : " + Saturation.ToString());
					writer.WriteLine("Auto Shutter     : " + AutoShutter);
					if (!AutoShutter)
						writer.WriteLine("Shutter          : " + Shutter.ToString());
					writer.WriteLine("Auto WB          : " + AutoWhiteBalance);
					if (!AutoWhiteBalance)
					{
						writer.WriteLine("WB-Blue          : " + WhiteBalanceBlue.ToString());
						writer.WriteLine("WB-Red           : " + WhiteBalanceRed.ToString());
					}
					writer.WriteLine();
				}
				
				if (CaptureSource == VideoSource.Kinect)
				{
					writer.WriteLine("Microsoft Kinect");
					writer.WriteLine("----------------------------------------");
					writer.WriteLine("Streams          : " + Streams);
					if ((Streams & Streams.Image) == Streams.Image) writer.WriteLine("Image Mode       : " + ImageMode);
					if ((Streams & Streams.IR)    == Streams.IR)    writer.WriteLine("IR Mode          : " + IRMode);
					if ((Streams & Streams.Depth) == Streams.Depth) writer.WriteLine("Depth Mode       : " + DepthMode);
					writer.WriteLine("Sync Mode        : " + SyncMode);
					writer.WriteLine();
				}
			}
		}
	}
}
