﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: RendererForm.cs 535 2012-04-04 14:50:13Z cr333 $
*/
using System;
using System.ComponentModel;
using System.Diagnostics;
using System.IO;
using System.Threading;
using System.Windows.Forms;
using VideoLib.OpenGL;
using RGBZcamera.Library.Config;

namespace RGBZcamera.Renderer
{
	public partial class RendererForm : Form
	{
		public RendererForm()
		{
			InitializeComponent();
		}

		private void ReprojectForm_Load(object sender, EventArgs eventArgs)
		{
			// List all data sets that can be found.
			PopulateVideoSourceComboBox();

			// Make vectors editable.
			TypeDescriptor.AddAttributes(typeof(OpenTK.Vector2), new TypeConverterAttribute(typeof(VectorConverter)));
			TypeDescriptor.AddAttributes(typeof(OpenTK.Vector3), new TypeConverterAttribute(typeof(VectorConverter)));
			TypeDescriptor.AddAttributes(typeof(OpenTK.Vector4), new TypeConverterAttribute(typeof(VectorConverter)));

			// Show intrinsic and extrinsic camera parameters.
			TypeDescriptor.AddAttributes(typeof(VideoLib.Vision.ExtrinsicParameters), new TypeConverterAttribute(typeof(ExpandableObjectConverter)));
			TypeDescriptor.AddAttributes(typeof(VideoLib.Vision.IntrinsicParameters), new TypeConverterAttribute(typeof(ExpandableObjectConverter)));
		}

		#region Populate source video list
//=================================================================================================

		private void PopulateVideoSourceComboBox()
		{
			videoSourceComboBox.Items.Clear();

			// Find videos in the the data directory ("/data/input/" in some parent directory).
			var dataDir = ConfigReader.FindDataDirectory(".");
			if (dataDir != null)
			{
				var dataInputDir = new DirectoryInfo(Path.Combine(dataDir.FullName, "input"));
				RecursivelyFindInputVideos(dataInputDir);
			}

			if (videoSourceComboBox.Items.Count == 0)
			{
				// Disable the combo box if no videos are found.
				videoSourceComboBox.Items.Add("No videos found.");
				videoSourceComboBox.SelectedIndex = 0;
				videoSourceComboBox.Enabled = false;
				startVideoButton.Enabled    = false;
			}
			else
			{
				// Select the first video.
				videoSourceComboBox.SelectedIndex = 0;
			}
		}

		/// <summary>Recursively searches a directory tree for config files describing input videos.</summary>
		/// <param name="dataDir">Directory to search in.</param>
		private void RecursivelyFindInputVideos(DirectoryInfo dataDir)
		{
			// Ignore invalid data directories.
			if (dataDir == null)
				return;

			// Check for input video in this directory.
			var configFile = new FileInfo(Path.Combine(dataDir.FullName, "_config.ini"));
			if (configFile.Exists)
				videoSourceComboBox.Items.Add(new InputVideo(configFile));

			// Find input videos recursively.
			foreach (var dir in dataDir.GetDirectories())
				RecursivelyFindInputVideos(dir);
		}

//=================================================================================================
		#endregion Populate source video list

		#region Style or Node selection
//=================================================================================================

		public VideoLib.Filters.SwitchFilter StyleSelector { get; set; }

		private void meshRadioButton_CheckedChanged(       object sender, EventArgs e) { if (meshRadioButton.Checked)        SelectStyleNode(0); }
		private void debugRadioButton_CheckedChanged(      object sender, EventArgs e) { if (debugRadioButton.Checked)       SelectStyleNode(1); }
		private void abstractionRadioButton_CheckedChanged(object sender, EventArgs e) { if (abstractionRadioButton.Checked) SelectStyleNode(2); }
		private void relightingRadioButton_CheckedChanged( object sender, EventArgs e) { if (relightingRadioButton.Checked)  SelectStyleNode(3); }
		private void hatchingRadioButton_CheckedChanged(   object sender, EventArgs e) { if (hatchingRadioButton.Checked)    SelectStyleNode(5); }
		private void backgroundRadioButton_CheckedChanged( object sender, EventArgs e) { if (backgroundRadioButton.Checked)  SelectStyleNode(7); }
		private void stereoRadioButton_CheckedChanged(     object sender, EventArgs e) { if (stereoRadioButton.Checked)      SelectStyleNode(8); }

		private void alignmentRadioButton_CheckedChanged(  object sender, EventArgs e) { if (alignmentRadioButton.Checked)   SelectFirstMatchingNode(typeof(RGBZcamera.Library.ReprojectionNode)); }
		private void fillInRadioButton_CheckedChanged(     object sender, EventArgs e) { if (fillInRadioButton.Checked)      SelectFirstMatchingNode(typeof(RGBZcamera.Library.FillIn.MultiResolutionFillIn)); }
		private void filterRadioButton_CheckedChanged(     object sender, EventArgs e) { if (filterRadioButton.Checked)      SelectFirstMatchingNode(typeof(RGBZcamera.Library.SpatioTemporalFilter)); }
		private void styleRadioButton_CheckedChanged(      object sender, EventArgs e) { if (styleRadioButton.Checked)       SelectStyleNode(); }

		/// <summary>Shows the node corresponding to the currently-selected style in the property grid.</summary>
		public void SelectStyleNode()
		{
			if (StyleSelector == null)
				return;

			var type = ConvertStyleToType(StyleSelector.SelectedOutput);

			if(type == null)
				return;

			SelectFirstMatchingNode(type);
		}

		/// <summary>Selects a particular style and shows its node in the property grid.</summary>
		private void SelectStyleNode(int style)
		{
			// set new rendering style
			if (StyleSelector != null)
				StyleSelector.SelectedOutput = style;

			// select rendering node
			SelectStyleNode();
		}

		private Type ConvertStyleToType(int style)
		{
			switch (style)
			{
				case 0:  return typeof(RGBZcamera.Library.MeshRenderer);
				case 1:  return typeof(RGBZcamera.Library.AbstractedLines.ToneRendererNode);
				case 2:  return typeof(RGBZcamera.Library.AbstractedLines.AbstractedLinesFilter);
				case 3:  return typeof(RGBZcamera.Library.Relighter);
				case 4:  return typeof(RGBZcamera.Library.XToon.XToonShader);
				case 5:  return typeof(RGBZcamera.Library.Hatching.HatchingFilter);
				case 6:  return typeof(RGBZcamera.Library.WinnemoellerAbstraction.WinnemoellerAbstraction);
				case 7:  return typeof(RGBZcamera.Library.BackgroundSubtraction);
				case 8:  return typeof(RGBZcamera.Library.StereoWarper);
				default: return null;
			}			
		}

		/// <summary>Selects the first node of matching type.</summary>
		/// <param name="type">The node type.</param>
		private void SelectFirstMatchingNode(Type type)
		{
			foreach (var item in nodeComboBox.Items)
			{
				if (item.GetType().Equals(type))
				{
					nodeComboBox.SelectedItem = item;
					UpdateStageButtons(type);
					return;
				}
			}

			Debug.WriteLine(string.Format("Node of type {0} not found in nodeComboBox.", type));
		}

		private void UpdateStageButtons(Type type)
		{
			alignmentRadioButton.Checked = (type == typeof(RGBZcamera.Library.ReprojectionNode));
			fillInRadioButton.Checked    = (type == typeof(RGBZcamera.Library.FillIn.MultiResolutionFillIn));
			filterRadioButton.Checked    = (type == typeof(RGBZcamera.Library.SpatioTemporalFilter));
			styleRadioButton.Checked     = (StyleSelector != null && type == ConvertStyleToType(StyleSelector.SelectedOutput));
		}

//=================================================================================================
		#endregion Style or Node selection

		#region Start/Stop video playback
//=================================================================================================

		private void startVideoButton_Click(object sender, EventArgs e)
		{
			// Read out the options from the GUI.
			var video = videoSourceComboBox.SelectedItem as InputVideo;
			var settings = new RendererSettings(video.Config)
			{
				Loop                = loopCheckBox.Enabled                && loopCheckBox.Checked,
				ToCache             = toCacheCheckBox.Enabled             && toCacheCheckBox.Checked,
				FromCache           = fromCacheCheckBox.Enabled           && fromCacheCheckBox.Checked,
				SaveRenderings      = saveRenderingsCheckBox.Enabled      && saveRenderingsCheckBox.Checked,
				DisplayRenderings   = displayRenderingsCheckBox.Enabled   && displayRenderingsCheckBox.Checked,
				DisplayDebugOutputs = displayDebugOutputsCheckBox.Enabled && displayDebugOutputsCheckBox.Checked,
			};

			// Disable controls.
			videoSourceComboBox.Enabled         = false;
			loopCheckBox.Enabled                = false;
			startVideoButton.Enabled            = false;
			toCacheCheckBox.Enabled             = false;
			fromCacheCheckBox.Enabled           = false;
			saveRenderingsCheckBox.Enabled      = false;
			displayRenderingsCheckBox.Enabled   = false;
			displayDebugOutputsCheckBox.Enabled = false;
			alignmentRadioButton.Enabled        = !settings.FromCache;
			fillInRadioButton.Enabled           = !settings.FromCache;
			filterRadioButton.Enabled           = !settings.FromCache;

			// Enable controls
			stopVideoButton.Enabled = true;
			abstractionRadioButton.Checked = true;

			// Show controls
			mainTableLayoutPanel.SuspendLayout();
			settingsGroupBox.Visible              = true;
			inputVideoAndDepthGroupBox.Visible    = settings.DisplayDebugOutputs;
			alignedAndCroppedGroupBox.Visible     = settings.DisplayDebugOutputs;
			filteredDepthGroupBox.Visible         = settings.DisplayDebugOutputs;
			renderingOutputGroupBox.Visible       = settings.DisplayRenderings;
			mainTableLayoutPanel.ResumeLayout();

			// Start the video with the given settings.
			new Thread(delegate() { StartVideo(settings); }).Start();
		}

		private void stopVideoButton_Click(object sender, EventArgs e)
		{
			new Thread(StopVideo).Start();
		}

		// NB: blocks
		private void StartVideo(RendererSettings settings)
		{
			try
			{
				RendererProgram.Start(this, settings);
			}
			catch (Exception e)
			{
				MessageBox.Show(
					String.Format("{0}: {1}\n\n{2}\n\n(RendererForm.StartVideo)", e.GetType().Name, e.Message, e),
					String.Format("Error: {0}", e.GetType()), MessageBoxButtons.OK, MessageBoxIcon.Error);
				Application.Exit();
				return;
			}

			if (Created && !Disposing)
				StopVideo();
		}
		
		private void StopVideo()
		{
			if(RendererProgram.scheduler != null) RendererProgram.scheduler.Stop();
			if(RendererProgram.topology  != null) RendererProgram.topology.Clear();
			RendererProgram.scheduler = null;
			RendererProgram.topology  = null;

			this.BeginInvoke(new Action(delegate()
			{
				//nodeComboBox.AutoCompleteSource = AutoCompleteSource.None;
				nodeComboBox.SelectedItem = null;

				// Disable controls.
				stopVideoButton.Enabled = false;
				
				// Hide controls.
				mainTableLayoutPanel.SuspendLayout();
				settingsGroupBox.Visible            = false;
				inputVideoAndDepthGroupBox.Visible  = false;
				alignedAndCroppedGroupBox.Visible   = false;
				filteredDepthGroupBox.Visible       = false;
				renderingOutputGroupBox.Visible     = false;
				mainTableLayoutPanel.ResumeLayout();
				
				// Enable controls (depending on settings :s).
				videoSourceComboBox.Enabled         = true;
				loopCheckBox.Enabled                = !toCacheCheckBox.Checked || !saveRenderingsCheckBox.Checked;
				startVideoButton.Enabled            = true;
				toCacheCheckBox.Enabled             = !fromCacheCheckBox.Checked;
				fromCacheCheckBox.Enabled           = !toCacheCheckBox.Checked;
				saveRenderingsCheckBox.Enabled      = true;
				displayRenderingsCheckBox.Enabled   = true;
				displayDebugOutputsCheckBox.Enabled = !fromCacheCheckBox.Checked;
			}));
		}

//=================================================================================================
		#endregion Start/Stop video playback

		#region GUI logic
//=================================================================================================
		
		private void videoSourceComboBox_SelectedIndexChanged(object sender, EventArgs e)
		{
			var video = videoSourceComboBox.SelectedItem as InputVideo;
			if (video == null) return;

			// Enable/disable controls as appropriate.
			loopCheckBox.Enabled                = video.DataFound;
			startVideoButton.Enabled            = video.DataFound;
			toCacheCheckBox.Enabled             = video.DataFound;
			fromCacheCheckBox.Enabled           = video.CacheFound;
			saveRenderingsCheckBox.Enabled      = video.DataFound;
			displayRenderingsCheckBox.Enabled   = video.DataFound;
			displayDebugOutputsCheckBox.Enabled = video.DataFound;
		}

		private void toCacheCheckBox_CheckedChanged(object sender, EventArgs e)
		{
			// Disable looping when writing processed videos to cache.
			loopCheckBox.Enabled = !toCacheCheckBox.Checked;

			// Prevent loading from and writing to cache at the same time.
			fromCacheCheckBox.Enabled = !toCacheCheckBox.Checked;
		}

		private void fromCacheCheckBox_CheckedChanged(object sender, EventArgs e)
		{
			// Cannot show debug output when loading the processing video from cache.
			displayDebugOutputsCheckBox.Enabled = !fromCacheCheckBox.Checked;

			// Prevent loading from and writing to cache at the same time.
			toCacheCheckBox.Enabled = !fromCacheCheckBox.Checked;
		}

		private void saveRenderingsCheckBox_CheckedChanged(object sender, EventArgs e)
		{
			// Disable looping when writing renderings to disk.
			loopCheckBox.Enabled = !saveRenderingsCheckBox.Checked;
		}

		private void nodeComboBox_SelectedIndexChanged(object sender, EventArgs e)
		{
			// Show the selected item in the property grid to edit its properties.
			nodePropertyGrid.SelectedObject = nodeComboBox.SelectedItem;

			// Additionally indicate which processing stage this node corresponds to.
			if (nodeComboBox.SelectedItem != null)
				UpdateStageButtons(nodeComboBox.SelectedItem.GetType());
		}

		private void contextMenu_Opening(object sender, CancelEventArgs e)
		{
			// Disable "Reset" item if not possible (also if already set to default value).
			resetItem.Enabled = nodePropertyGrid.SelectedGridItem.PropertyDescriptor.CanResetValue(nodePropertyGrid.SelectedObject);
		}

		private void resetItem_Click(object sender, EventArgs e)
		{
			// Reset value of selected property and refresh the grid.
			nodePropertyGrid.SelectedGridItem.PropertyDescriptor.ResetValue(nodePropertyGrid.SelectedObject);
			nodePropertyGrid.Refresh();
		}

		private void videoControl_DoubleClick(object sender, EventArgs e)
		{
			SelectVideoDisplay(sender);
		}

		/// <summary>Selects the video display corresponding to a given VideoControl.</summary>
		/// <param name="videoDisplay">The video control.</param>
		private void SelectVideoDisplay(object videoDisplay)
		{
			foreach (var item in nodeComboBox.Items)
			{
				var display = item as VideoDisplay;
				if (display == null) continue;

				if (display.VideoControl.Equals(videoDisplay))
					nodeComboBox.SelectedItem = display;
			}
		}

//=================================================================================================
		#endregion GUI logic
	}
}
