﻿// $Id: StereoVideoControl.cs 489 2010-11-09 18:45:40Z cr333 $
using System;
using System.Diagnostics;
using System.Drawing;
using System.Windows.Forms;
using OpenTK;
using OpenTK.Graphics.OpenGL;
using VideoLib.Frames;
using VideoLib.Sinks;
//using OpenTK.Graphics;

namespace VideoLib.OpenGL
{
	// TODO: need to free texture when disposing

	/// <summary>Implements an OpenGL-based video rendering control.</summary>
	public partial class StereoVideoControl : GLControl, IStereoVideoControl
	{
		/// <summary>Handle to the texture that will hold the left video frames.</summary>
		private int textureIdLeft = 0;
		
		/// <summary>Handle to the texture that will hold the right video frames.</summary>
		private int textureIdRight = 0;

		/// <summary>Copy of the latest caption.</summary>
		private string caption = "";

		/// <summary>The size of the allocated texture.</summary>
		Size textureSize = Size.Empty;

		/// <summary>Gets or sets a value indicating whether to show the caption and FPS counter.</summary>
		public bool ShowCaption { get; set; }

		/// <summary>Indicates if the control is currently displayed in the designer.</summary>
		protected bool InDesignMode
		{
			get { if (DesignMode) return true; return this.ParentForm.Site != null; }
		}

		/// <summary>Constructs an OpenGL-based video rendering control.</summary>
		public StereoVideoControl()
			: base (new OpenTK.Graphics.GraphicsMode(new OpenTK.Graphics.ColorFormat(32), 8, 0, 0, new OpenTK.Graphics.ColorFormat(32), 2, true))
		{
			InitializeComponent();

			// set custom style
			this.SetStyle(ControlStyles.Selectable, false);
			this.SetStyle(ControlStyles.ResizeRedraw, true);

			// set default property values
			ShowCaption = false;
		}

		/// <summary>Initialises the OpenGL state for this control.</summary>
		private void InitGL()
		{
			GL.ClearColor(Color.SkyBlue);
			GL.Enable(EnableCap.Texture2D);
			GL.Disable(EnableCap.Blend);
			GL.Disable(EnableCap.DepthTest);
			GL.Disable(EnableCap.Lighting);
			GL.DepthMask(false);
		}

		/// <summary>Sets up the canonical viewport.</summary>
		private void SetupViewport()
		{
			MakeCurrent();

			GL.Viewport(0, 0, Width, Height); // Use all of the glControl painting area

			GL.MatrixMode(MatrixMode.Projection);
			GL.LoadIdentity();
			GL.Ortho(0, Width, 0, Height, -1, 1); // Bottom-left corner pixel has coordinate (0, 0)

			GL.MatrixMode(MatrixMode.Modelview);
			GL.LoadIdentity();
		}

		/// <summary>Allocates a texture of the desired size.</summary>
		/// <param name="size">Texture size.</param>
		private void AllocTextures(Size size)
		{
			// skip if the texture already has the desired size
			if (textureSize == size) return;

			// free any existing texture
			FreeTextures();

			// generate a new texture, bind to it and fill it with darkness
			textureSize = size;

			textureIdLeft = GL.GenTexture();
			GL.BindTexture(TextureTarget.Texture2D, textureIdLeft);
			GL.TexImage2D(TextureTarget.Texture2D, 0, PixelInternalFormat.Rgba, size.Width, size.Height, 0, OpenTK.Graphics.OpenGL.PixelFormat.Rgb, PixelType.UnsignedByte, IntPtr.Zero);
			GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMinFilter, (int)TextureMinFilter.Linear);
			GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMagFilter, (int)TextureMagFilter.Linear);
			GL.BindTexture(TextureTarget.Texture2D, 0);

			textureIdRight = GL.GenTexture();
			GL.BindTexture(TextureTarget.Texture2D, textureIdRight);
			GL.TexImage2D(TextureTarget.Texture2D, 0, PixelInternalFormat.Rgba, size.Width, size.Height, 0, OpenTK.Graphics.OpenGL.PixelFormat.Rgb, PixelType.UnsignedByte, IntPtr.Zero);
			GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMinFilter, (int)TextureMinFilter.Linear);
			GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMagFilter, (int)TextureMagFilter.Linear);
			GL.BindTexture(TextureTarget.Texture2D, 0);
		}

		/// <summary>Frees the allocated texture.</summary>
		private void FreeTextures()
		{
			if (textureIdLeft != 0)
			{
				GL.DeleteTexture(textureIdLeft);
				textureIdLeft = 0;
			}
			if (textureIdRight != 0)
			{
				GL.DeleteTexture(textureIdRight);
				textureIdRight = 0;
			}
		}

		/// <summary>Accepts a new frame and caption to be drawn onto this control.</summary>
		/// <param name="bitmapFrameL">The new left frame.</param>
		/// <param name="bitmapFrameR">The new right frame.</param>
		/// <param name="caption">The new caption (ignored for now).</param>
		public void Draw(BitmapFrame bitmapFrameL, BitmapFrame bitmapFrameR, string caption)
		{
			// only go ahead if it is safe to do so
			if (!Disposing && Created)
			{
				// need to draw the image on the UI thread
				BeginInvoke(new Action(delegate() { DrawImageInternal(bitmapFrameL, bitmapFrameR, caption); }));
			}
		}

		/// <summary>Accepts a new image and caption to be drawn onto this control.</summary>
		/// <remarks>Needs to run on the GUI/OpenGL thread.</remarks>
		/// <param name="bitmapL">The new left image.</param>
		/// <param name="bitmapR">The new right image.</param>
		/// <param name="caption">The new caption (ignored for now).</param>
		private void DrawImageInternal(BitmapFrame bitmapL, BitmapFrame bitmapR, string caption)
		{
			// this handles most crashes when closing the form
			if (Disposing || !Created)
				return;

			this.caption = caption;

			Size bitmapSize = new Size(bitmapL.Width, bitmapL.Height);

			// implement auto-sizing
			if (AutoSize && textureSize != bitmapSize)
			{
				this.Size = bitmapSize;
				MinimumSize = bitmapSize;
				MaximumSize = bitmapSize;
			}

			// make sure we've got a texture of the right size
			AllocTextures(bitmapSize);

			// replace texture contents with new frame data

			FillTexture(textureIdLeft, bitmapL);
			FillTexture(textureIdRight, bitmapR);

			// set up the viewport and then ask for a full redraw
			SetupViewport();
			Invalidate();
		}

		private void FillTexture(int textureId, BitmapFrame bitmap)
		{
			GL.BindTexture(TextureTarget.Texture2D, textureId);

			switch (bitmap.Format)
			{
				case VideoLib.Frames.PixelFormat.Grey8:
					GL.TexSubImage2D(TextureTarget.Texture2D, 0, 0, 0, bitmap.Width, bitmap.Height,
						OpenTK.Graphics.OpenGL.PixelFormat.Luminance, PixelType.UnsignedByte, bitmap.Data);
					break;

				case VideoLib.Frames.PixelFormat.BGR24:
					GL.PixelStore(PixelStoreParameter.UnpackAlignment, 1);
					GL.TexSubImage2D(TextureTarget.Texture2D, 0, 0, 0, bitmap.Width, bitmap.Height,
						OpenTK.Graphics.OpenGL.PixelFormat.Bgr, PixelType.UnsignedByte, bitmap.Data);
					break;

				case VideoLib.Frames.PixelFormat.BGR32:
				case VideoLib.Frames.PixelFormat.BGRA32:
					GL.PixelStore(PixelStoreParameter.UnpackAlignment, 4);
					GL.TexSubImage2D(TextureTarget.Texture2D, 0, 0, 0, bitmap.Width, bitmap.Height,
						OpenTK.Graphics.OpenGL.PixelFormat.Bgra, PixelType.UnsignedByte, bitmap.Data);
					break;

				case VideoLib.Frames.PixelFormat.Float:
					GL.PixelStore(PixelStoreParameter.UnpackAlignment, 4);
					GL.TexSubImage2D(TextureTarget.Texture2D, 0, 0, 0, bitmap.Width, bitmap.Height,
						OpenTK.Graphics.OpenGL.PixelFormat.Luminance, PixelType.Float, bitmap.Data);
					break;

				default:
					// previously, this threw an ArgumentException, but it can't be handled properly, so just stay black
					break;
			}
			GL.BindTexture(TextureTarget.Texture2D, 0);
		}

		/// <summary>Renders the current video frame.</summary>
		private void Render()
		{
			if (textureIdLeft != 0 && textureIdRight != 0)
			{
				GL.DrawBuffer(DrawBufferMode.BackLeft);
				GL.BindTexture(TextureTarget.Texture2D, textureIdLeft);
				GL.Begin(BeginMode.Quads); // NB: texcords apply y-flip here
				GL.TexCoord2(0, 1); GL.Vertex2(0, 0);
				GL.TexCoord2(1, 1); GL.Vertex2(Width, 0);
				GL.TexCoord2(1, 0); GL.Vertex2(Width, Height);
				GL.TexCoord2(0, 0); GL.Vertex2(0, Height);
				GL.End();
				Debug.Assert(GL.GetError() == ErrorCode.NoError);
				GL.BindTexture(TextureTarget.Texture2D, 0);


				if (ShowCaption)
				{
					using (OpenTK.Graphics.TextPrinter textPrinter = new OpenTK.Graphics.TextPrinter())
					{
						textPrinter.Begin();
						textPrinter.Print(caption, Font, Color.LightGreen, new RectangleF(2, 2, Width - 2, Height - 2));
						textPrinter.End();
					}
				}

				GL.DrawBuffer(DrawBufferMode.BackRight);
				GL.BindTexture(TextureTarget.Texture2D, textureIdRight);
				GL.Begin(BeginMode.Quads); // NB: texcords apply y-flip here
				GL.TexCoord2(0, 1); GL.Vertex2(0, 0);
				GL.TexCoord2(1, 1); GL.Vertex2(Width, 0);
				GL.TexCoord2(1, 0); GL.Vertex2(Width, Height);
				GL.TexCoord2(0, 0); GL.Vertex2(0, Height);
				GL.End();
				Debug.Assert(GL.GetError() == ErrorCode.NoError);
				GL.BindTexture(TextureTarget.Texture2D, 0);


				if (ShowCaption)
				{
					using (OpenTK.Graphics.TextPrinter textPrinter = new OpenTK.Graphics.TextPrinter())
					{
						textPrinter.Begin();
						textPrinter.Print(caption, Font, Color.LightGreen, new RectangleF(2, 2, Width - 2, Height - 2));
						textPrinter.End();
					}
				}
			}
			else
			{
				GL.Clear(ClearBufferMask.ColorBufferBit);
			}

			// show results
			SwapBuffers();
		}

		#region Event handling

		private void StereoVideoControl_Load(object sender, EventArgs e)
		{
			if (!InDesignMode)
				InitGL();
		}

		private void StereoVideoControl_Paint(object sender, PaintEventArgs e)
		{
			// custom film strip look for Designer
			if (InDesignMode)
			{
				e.Graphics.Clear(BackColor);
				int framePadding = 4;
				int holePadding = 6;
				int holeSize = 10;

				// black background
				e.Graphics.FillRectangle(new SolidBrush(Color.Black), ClientRectangle);

				// punched holes
				for (int y = 4; y < Height; y += 2 * holeSize)
				{
					e.Graphics.FillRectangle(new SolidBrush(Color.LightGray), holePadding, y, holeSize, holeSize);
					e.Graphics.FillRectangle(new SolidBrush(Color.LightGray), Width - holePadding - holeSize, y, holeSize, holeSize);
				}

				// frames
				int frameBoundary1 = (int)(0.1 * Height + 0.5);
				int frameBoundary2 = (int)(0.9 * Height + 0.5);
				e.Graphics.FillRectangle(new SolidBrush(Color.LightGray), 2 * holePadding + holeSize, 0, Width - 4 * holePadding - 2 * holeSize, frameBoundary1 - framePadding / 2);
				e.Graphics.FillRectangle(new SolidBrush(Color.LightGray), 2 * holePadding + holeSize, frameBoundary1 + framePadding / 2, Width - 4 * holePadding - 2 * holeSize, frameBoundary2 - frameBoundary1 - framePadding);
				e.Graphics.FillRectangle(new SolidBrush(Color.LightGray), 2 * holePadding + holeSize, frameBoundary2 + framePadding / 2, Width - 4 * holePadding - 2 * holeSize, Height - (frameBoundary2 + framePadding / 2));

				// "Video Panel" text
				Font font = new Font(Font.FontFamily, 16, FontStyle.Bold);
				String text = "OpenGL" + Environment.NewLine + "Video" + Environment.NewLine + "Panel";
				StringFormat stringFormat = new StringFormat { Alignment = StringAlignment.Center, LineAlignment = StringAlignment.Center };
				e.Graphics.DrawString(text, font, new SolidBrush(ForeColor), Width / 2, Height / 2, stringFormat);
			}
			else
			{
				MakeCurrent();
				Render();
			}
		}

		#endregion
	}
}
