﻿// $Id: VideoControl.cs 545 2011-12-07 17:39:30Z cr333 $
using System;
using System.Diagnostics;
using System.Drawing;
using System.Windows.Forms;
using OpenTK;
using OpenTK.Graphics.OpenGL;
using VideoLib.Frames;
using VideoLib.Sinks;
using System.ComponentModel;

namespace VideoLib.OpenGL
{
	// TODO: need to free texture when disposing

	/// <summary>Enumerates the colour channel(s) to display.</summary>
	[Flags()]
	public enum DisplayOptions
	{
	//---- basic values ---------------------------------------------------------------------------

		//Nothing = 0,

		/// <summary>Display the red channel.</summary>
		R = 1,

		/// <summary>Display the green channel.</summary>
		G = 2,

		/// <summary>Display the blue channel.</summary>
		B = 4,

		/// <summary>Display the alpha channel.</summary>
		A = 8,

		/// <summary>Display selected channel as grey.</summary>
		AsGrey = 16,

	//---- composite values -----------------------------------------------------------------------

		/// <summary>Display the red and green channels.</summary>
		RG = R | G,

		/// <summary>Display the red and blue channels.</summary>
		RB = R | B,

		/// <summary>Display the green and blue channels.</summary>
		GB = G | B,

		/// <summary>Display the red, green and blue channels.</summary>
		RGB = R | G | B,

		/// <summary>Displays all channels (red, green, blue and alpha).</summary>
		RGBA = R | G | B | A,

		/// <summary>Display the red channel as grey.</summary>
		GreyR = R | AsGrey,

		/// <summary>Display the green channel as grey.</summary>
		GreyG = G | AsGrey,

		/// <summary>Display the blue channel as grey.</summary>
		GreyB = B | AsGrey,

		/// <summary>Display the alpha channel as grey.</summary>
		GreyA = A | AsGrey,
	}

	/// <summary>Implements an OpenGL-based video rendering control.</summary>
	public partial class VideoControl : GLControl, IVideoControl
	{
		/// <summary>Indicates if the control is currently displayed in the designer.</summary>
		protected bool InDesignMode
		{
			get { if (DesignMode) return true; return this.ParentForm.Site != null; }
		}

		/// <summary>Constructs an OpenGL-based video rendering control.</summary>
		public VideoControl()
		{
			InitializeComponent();

			// set custom style
			this.SetStyle(ControlStyles.Selectable, false);
			this.SetStyle(ControlStyles.ResizeRedraw, true);

			// set default property values
			PreserveAspectRatio = true;
			LinearInterpolation = true;
			ShowCaption         = false;
			Zoom                = 1.0f;

#if DEBUG
			Background = Color.Black;
#else
			Background = Color.FromKnownColor(KnownColor.Control);
#endif
		}

//=================================================================================================
		#region Fields and properties

		/// <summary>Gets or sets whether the aspect ratio of the input is preserved (false means stretch to fill). Default: true.</summary>
		[DefaultValue(true), Description("Indicates whether the aspect ratio of the input is preserved.")]
		public bool PreserveAspectRatio { get; set; }

		/// <summary>Gets or sets the zoom level.</summary>
		[DefaultValue(1.0f), Description("The zoom level applied to the display.")]
		public float Zoom { get; set; }

		/// <summary>Gets or sets the background colour behind and surrounding the video.</summary>
		[Description("The background colour behind and surrounding the video.")]
		public Color Background { get; set; }

		/// <summary>Gets or sets whether to show the caption and FPS counter.</summary>
		[DefaultValue(false), Description("Indicates whether to show the caption and a framerate counter.")]
		public bool ShowCaption { get; set; }

		/// <summary>Gets or sets whether linear interpolation is enabled.</summary>
		[DefaultValue(true), Description("Enables linear interpolation of the displayed texture.")]
		public bool LinearInterpolation { get; set; }

		/// <summary>Gets or sets the colour channel(s) to display.</summary>
		[DefaultValue(DisplayOptions.RGBA), Description("Colour channel(s) to display.")]
		public DisplayOptions Display { get { return display; } set { display = value; dirty = true; } }

		/// <summary>Gets or sets the scaling factor for pixel values.</summary>
		[DefaultValue(1.0f), Description("Pixel values are scaled with this factor.")]
		public float Factor { get { return factor; } set { factor = value; dirty = true; } }

		/// <summary>Gets or sets the offset applied to pixel values after scaling (pixel values are 0..1).</summary>
		[DefaultValue(0.0f), Description("The offset applied to pixel values after scaling (pixel values are 0..1).")]
		public float Offset { get { return offset; } set { offset = value; dirty = true; } }

		/// <summary>The texture that will hold the video frames.</summary>
		private Texture texture;

		/// <summary>Holds incoming <see cref="TextureFrame"/>s.</summary>
		private TextureFrame textureFrame;

		/// <summary>Copy of the latest caption.</summary>
		private string caption = "";

		/// <summary>Colour channels to display.</summary>
		private DisplayOptions display = DisplayOptions.RGBA;

		// private fields for shader uniforms
		private float factor = 1;
		private float offset = 0;

		#endregion // Fields and properties
//=================================================================================================
		#region Rendering functions

		private int programID;
		private bool dirty = true;

		/// <summary>Initialises the OpenGL state for this control.</summary>
		private void InitGL()
		{
			// load, compile and link shaders
			programID = Utils.LinkShaders(
				Utils.CreateShaderFromResource("VideoLib.OpenGL.Shaders.VideoControl.frag")
			);

			// set constant uniforms
			GL.UseProgram(programID);
			GL.Uniform1(GL.GetUniformLocation(programID, "input"), 0); // texture unit 0
			GL.UseProgram(0);
		}

		/// <summary>Accepts a new frame and caption to be drawn onto this control.</summary>
		/// <param name="frame">The new frame.</param>
		/// <param name="caption">The new caption (ignored for now).</param>
		public void Draw(Frame frame, string caption)
		{
			// only go ahead if it is safe to do so
			if (!Disposing && Created)
			{
				var textureFrame = frame as TextureFrame;
				if (textureFrame != null)
				{
					// hold onto frame
					textureFrame.RefInc();

					// need to draw the image on the UI thread
					BeginInvoke(new Action(delegate() { DrawImageInternal(textureFrame, caption); }));
				}
				else if (frame is BitmapFrame)
				{
					// need to draw the image on the UI thread
					BeginInvoke(new Action(delegate() { DrawImageInternal((BitmapFrame)frame, caption); }));
				}
			}
		}

		/// <summary>Accepts a new image and caption to be drawn onto this control.</summary>
		/// <remarks>Needs to run on the GUI/OpenGL thread.</remarks>
		/// <param name="bitmap">The new image.</param>
		/// <param name="caption">The new caption (ignored for now).</param>
		private void DrawImageInternal(BitmapFrame bitmap, string caption)
		{
			// this handles most crashes when closing the form
			if (Disposing || !Created)
				return;

			this.caption = caption;

			// implement auto-sizing
			if (AutoSize)
			{
				Size bitmapSize = new Size((int)(Zoom * bitmap.Width + 0.5f), (int)(Zoom * bitmap.Height + 0.5f));
				this.Size   = bitmapSize;
				MinimumSize = bitmapSize;
				MaximumSize = bitmapSize;
			}

			MakeCurrent();

			// allocate new texture on startup
			if (texture == null)
			{
				texture = new Texture(bitmap.Width, bitmap.Height)
				{
					Filter = (LinearInterpolation ? TextureMagFilter.Linear : TextureMagFilter.Nearest),
					Wrap = TextureWrapMode.ClampToEdge // avoid fade-to-black borders in low-res images
				};
			}

			// update the texture with new frame data
			texture.Update(bitmap);

			// ask for a full redraw
			Invalidate();
		}

		/// <summary>Accepts a new image and caption to be drawn onto this control.</summary>
		/// <remarks>Needs to run on the GUI/OpenGL thread.</remarks>
		/// <param name="bitmap">The new image.</param>
		/// <param name="caption">The new caption (ignored for now).</param>
		private void DrawImageInternal(TextureFrame bitmap, string caption)
		{
			// this handles most crashes when closing the form
			if (Disposing || !Created)
				return;

			this.caption = caption;

			// implement auto-sizing
			if (AutoSize)
			{
				Size bitmapSize = new Size((int)(Zoom * bitmap.Width + 0.5f), (int)(Zoom * bitmap.Height + 0.5f));
				this.Size   = bitmapSize;
				MinimumSize = bitmapSize;
				MaximumSize = bitmapSize;
			}

			MakeCurrent();

			if(textureFrame != null)
				textureFrame.RefDec();

			textureFrame = bitmap;

			// ask for a full redraw
			Invalidate();
		}

		/// <summary>Renders the current video frame.</summary>
		private void Render()
		{
			// set up matrices and viewport
			GL.MatrixMode(MatrixMode.Projection); GL.LoadIdentity();
			GL.MatrixMode(MatrixMode.Modelview);  GL.LoadIdentity();
			GL.Viewport(0, 0, Width, Height);

			// first clear the background, regardless of whether we have data
			GL.ClearColor(Background);
			GL.Clear(ClearBufferMask.ColorBufferBit);

			if (texture != null || textureFrame != null)
			{
				// texture to use
				var tex = (textureFrame != null ? textureFrame.Texture : texture);

				// necessary setup for texturing without depth
				GL.Enable(EnableCap.Texture2D);
				//GL.Disable(EnableCap.Blend); // ?
				GL.Disable(EnableCap.DepthTest);
				//GL.Disable(EnableCap.Lighting); // ?
				GL.DepthMask(false);

				float w, h;
				ComputeTargetSize(tex, out w, out h);

				GL.UseProgram(programID);
				tex.Bind();

				if (dirty)
				{
					GL.Uniform1(GL.GetUniformLocation(programID, "display"), (int)Display);
					GL.Uniform1(GL.GetUniformLocation(programID, "factor"), Factor);
					GL.Uniform1(GL.GetUniformLocation(programID, "offset"), Offset);
					dirty = false;
				}

				GL.Begin(BeginMode.Quads); // NB: texcords apply y-flip here
				GL.TexCoord2(0, 1); GL.Vertex2(-w, -h);
				GL.TexCoord2(1, 1); GL.Vertex2( w, -h);
				GL.TexCoord2(1, 0); GL.Vertex2( w,  h);
				GL.TexCoord2(0, 0); GL.Vertex2(-w,  h);
				GL.End();

				GL.UseProgram(0);
				tex.Unbind();

				if (ShowCaption)
				{
					using (OpenTK.Graphics.TextPrinter textPrinter = new OpenTK.Graphics.TextPrinter())
					{
						textPrinter.Begin();
						textPrinter.Print(caption, Font, Color.LightGreen, new RectangleF(2, 2, Width - 2, Height - 2));
						textPrinter.End();
					}
				}

				// restore depth buffer state
				GL.DepthMask(true);
				GL.Enable(EnableCap.DepthTest);
			}

			// show results
			SwapBuffers();
		}

		/// <summary>Fits the texture into the target width/height.</summary>
		/// <param name="tex">The input video texture.</param>
		/// <param name="targetWidth">Target Width.</param>
		/// <param name="targetHeight">Target height.</param>
		private void ComputeTargetSize(Texture tex, out float targetWidth, out float targetHeight)
		{
			targetWidth  = Width;
			targetHeight = Height;

			if (PreserveAspectRatio)
			{
				// fit to width if input image is wider than output image
				if (tex.Width * targetHeight >= targetWidth * tex.Height) // multiplied out numerator
					targetHeight = (float)tex.Height / tex.Width * targetWidth;
				else // fit to height if input image is higher than output image
					targetWidth = (float)tex.Width / tex.Height * targetHeight;
			}

			// normalise to get vertex coordinates
			targetWidth  /= Width;
			targetHeight /= Height;
		}

		#endregion // Rendering functions
//=================================================================================================
		#region Event handling + designer view

		private void VideoControl_Load(object sender, EventArgs e)
		{
			if (!InDesignMode)
				InitGL();
		}

		private void VideoControl_Paint(object sender, PaintEventArgs e)
		{
			// custom film strip look for Designer
			if (InDesignMode)
			{
				e.Graphics.Clear(BackColor);
				int framePadding = 4;
				int holePadding = 6;
				int holeSize = 10;

				// black background
				e.Graphics.FillRectangle(new SolidBrush(Color.Black), ClientRectangle);

				// punched holes
				for (int y = 4; y < Height; y += 2 * holeSize)
				{
					e.Graphics.FillRectangle(new SolidBrush(Color.LightGray), holePadding, y, holeSize, holeSize);
					e.Graphics.FillRectangle(new SolidBrush(Color.LightGray), Width - holePadding - holeSize, y, holeSize, holeSize);
				}

				// frames
				int frameBoundary1 = (int)(0.1 * Height + 0.5);
				int frameBoundary2 = (int)(0.9 * Height + 0.5);
				e.Graphics.FillRectangle(new SolidBrush(Color.LightGray), 2 * holePadding + holeSize, 0, Width - 4 * holePadding - 2 * holeSize, frameBoundary1 - framePadding / 2);
				e.Graphics.FillRectangle(new SolidBrush(Color.LightGray), 2 * holePadding + holeSize, frameBoundary1 + framePadding / 2, Width - 4 * holePadding - 2 * holeSize, frameBoundary2 - frameBoundary1 - framePadding);
				e.Graphics.FillRectangle(new SolidBrush(Color.LightGray), 2 * holePadding + holeSize, frameBoundary2 + framePadding / 2, Width - 4 * holePadding - 2 * holeSize, Height - (frameBoundary2 + framePadding / 2));

				// "Video Panel" text
				Font font = new Font(Font.FontFamily, 16, FontStyle.Bold);
				String text = "OpenGL" + Environment.NewLine + "Video" + Environment.NewLine + "Panel";
				StringFormat stringFormat = new StringFormat { Alignment = StringAlignment.Center, LineAlignment = StringAlignment.Center };
				e.Graphics.DrawString(text, font, new SolidBrush(ForeColor), Width / 2, Height / 2, stringFormat);
			}
			else
			{
				MakeCurrent();
				Render();
			}
		}

		#endregion // Event handling + designer view
//=================================================================================================
	}
}
