﻿// $Id: VideoDisplay.cs 498 2010-11-17 14:38:50Z cr333 $
using System;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using VideoLib.Frames;
using VideoLib.Utils;

namespace VideoLib.OpenGL
{
	/// <summary>Generic video display node with support for a caption.</summary>
	public class VideoDisplay : Node
	{
		/// <summary>Gets or sets the video panel.</summary>
		[Browsable(false)]
		public VideoControl VideoControl { get; set; } // TODO: handle Disposed event properly

		/// <summary>Gets or sets the caption to display in this panel.</summary>
		[Description("Caption of the displayed video.")]
		public string Caption { get; set; }

		/// <summary>For displaying the frames per second arriving at the display.</summary>
		protected FrameRateCounter fps = new FrameRateCounter(0.5, 1.0);
		
		/// <summary>Human-friendly node name.</summary>
		public override string Name
		{
			get { return String.Format("{0}({1})", base.Name, Caption); }
		}

		/// <summary>Indicates if the display rate should be limited to 25 fps, to avoid processing bottleneck.</summary>
		[DefaultValue(true), Description("Limits the display rate to 25 fps, to avoid processing bottleneck.")]
		public Boolean LimitDisplayRate { get; set; }
		private long nextFrame = 0;

//=================================================================================================
		#region Proxy properties

		/// <summary>Gets or sets whether the aspect ratio of the input is preserved (false means stretch to fill). Default: true.</summary>
		[DefaultValue(true), Description("Indicates whether the aspect ratio of the input is preserved.")]
		public bool PreserveAspectRatio { get { return VideoControl.PreserveAspectRatio; } set { if (VideoControl != null) VideoControl.PreserveAspectRatio = value; } }

		/// <summary>Gets or sets the zoom level to apply. Default: 1.</summary>
		[DefaultValue(1.0f), Description("The zoom level applied to the display.")]
		public float Zoom { get { return VideoControl.Zoom; } set { if (VideoControl != null) VideoControl.Zoom = value; } }

		/// <summary>Gets or sets the background colour behind and surrounding the video.</summary>
		[Description("The background colour behind and surrounding the video.")]
		public Color Background { get { return VideoControl.Background; } set { if (VideoControl != null) VideoControl.Background = value; } }

		/// <summary>Gets or sets whether to show the caption and FPS counter.</summary>
		[DefaultValue(false), Description("Indicates whether to show the caption and a framerate counter.")]
		public bool ShowCaption { get { return VideoControl.ShowCaption; } set { if (VideoControl != null) VideoControl.ShowCaption = value; } }

		/// <summary>Gets or sets whether linear interpolation is enabled.</summary>
		[DefaultValue(true), Description("Enables linear interpolation of the displayed texture.")]
		public bool LinearInterpolation { get { return VideoControl.LinearInterpolation; } set { if (VideoControl != null) VideoControl.LinearInterpolation = value; } }

		/// <summary>Gets or sets the colour channel(s) to display.</summary>
		[DefaultValue(DisplayOptions.RGBA), Description("Colour channel(s) to display.")]
		public DisplayOptions Display { get { return VideoControl.Display; } set { if (VideoControl != null) VideoControl.Display = value; } }

		/// <summary>Gets or sets the scaling factor for pixel values.</summary>
		[DefaultValue(1.0f), Description("Pixel values are scaled with this factor.")]
		public float Factor { get { return VideoControl.Factor; } set { if (VideoControl != null) VideoControl.Factor = value; } }

		/// <summary>Gets or sets the offset applied to pixel values after scaling (pixel values are 0..1).</summary>
		[DefaultValue(0.0f), Description("The offset applied to pixel values after scaling (pixel values are 0..1).")]
		public float Offset { get { return VideoControl.Offset; } set { if (VideoControl != null) VideoControl.Offset = value; } }

		#endregion // Fields and properties
//=================================================================================================

		/// <summary>Initialises a new instance of the <see cref="VideoDisplay"/> class.</summary>
		/// <param name="control">The target control.</param>
		public VideoDisplay(VideoControl control)
		{
			LimitDisplayRate = true;
			VideoControl = control;
			VideoControl.Disposed += new EventHandler(ControlDisposed);

			AddInputPin("in");
		}

		/// <summary>Initialises a new instance of the <see cref="VideoDisplay"/> class.</summary>
		/// <param name="control">The target control.</param>
		/// <param name="caption">The video caption.</param>
		public VideoDisplay(VideoControl control, string caption)
			: this(control)
		{
			Caption = caption;
		}

		/// <summary>Event handler for handling disposing of the video control.</summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
		void ControlDisposed(object sender, EventArgs e)
		{
			VideoControl = null;
			Status = NodeStatus.Stopped;
		}

		/// <summary>Processes a video frame.</summary>
		public override void Process()
		{
			// skip over this when not yet ready
			if (VideoControl == null) return;

			// get latest frame
			Frame frame = GetInputPin("in").GetFrame();
			fps.NewFrame();

			// stop at the end of the stream
			if (frame is EndOfStreamFrame)
			{
				Status = NodeStatus.Finished;
				return;
			}

			if (LimitDisplayRate && Stopwatch.GetTimestamp() < nextFrame)
			{
				// skip frame if necessary
			}
			else
			{
				// prepare caption
				string text = String.Format("{0}\nFrame: {1}\nFPS: {2:0.0}", Caption, frame.SequenceNumber, fps.FPS);

				// draw the new frame 
				VideoControl.Draw(frame, text);

				// compute time of next frame to display
				if(LimitDisplayRate)
					nextFrame = Stopwatch.GetTimestamp() + Stopwatch.Frequency / 25;
			}

			// decrement reference count, if applicable
			IRefCountedFrame refCounted = frame as IRefCountedFrame;
			if (refCounted != null) refCounted.RefDec();
		}
	}
}
