﻿// $Id: OpticalFlow.cs 493 2010-11-11 20:14:23Z cr333 $
#region Copyright (C) 2008 Martin Eisemann
/*
This source code is based on code by Martin Eisemann, released under GPL.
Ported to C# by Christian Richardt (University of Cambridge and MPI Informatik),
based on a Java port by James Tompkin (University College London).

This original C++ version of this source code is part of the Floating Textures test program,
which is a derivative of the complete Floating Textures program containing the offline version
with prereconstructed geometry and prerecorded video.

For more information on Floating Textures please have a look at the article

  "Floating Textures" 
  by M. Eisemann, B. De Decker, M. Magnor, P. Bekaert, E. de Aguiar, N. Ahmed, C. Theobalt, A. Sellent
  Computer Graphics Forum (Proc. Eurographics EG'08), vol. 27, no. 2, April 2008 

Copyright (C) 2008 Martin Eisemann

This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.
*/
#endregion

using System;
using System.Diagnostics;
using System.IO;
using System.Linq;
using OpenTK.Graphics.OpenGL;
using Color4 = OpenTK.Graphics.Color4;
using Glu = OpenTK.Graphics.Glu;
using System.Reflection;

// ignore 'Glu is obsolete' warning
#pragma warning disable 618

namespace VideoLib.OpenGL.OpticalFlow
{
	/// <summary>Calculates a variational optical flow between two or more images.</summary>
	/// <remarks>
	/// <para>
	/// GPU adapted implementation of the papers "High Accuracy Optical Flow Estimation
	/// Based on a Theory for Warping" by Brox, Bruhn, Papenberg and Weickert and
	/// "Towards Ultimate Motion Estimation: Combining Highest Accuracy with
	/// Real-Time Performance" by Bruhn and Weickert.
	/// </para>
	/// <para>
	/// Note that this is no exact implementation of the above mentioned papers
	/// but a GPU adapted version fitted to our needs for real-time processing.
	/// </para>
	/// <para>
	/// (Clarification of differences via email on 2010-10-27:
	///   (1) using Jacobi iterations instead of Gauss-Seidel (easier to parallelise);
	///   (2) computing forward and backward flow simultaneously;
	///   (3) computing half a V cycle instead of V or W cycle (only low res to high res).)
	/// </para>
	/// <para>
	/// Observations: A small coarseLevel leads to crispier results, but probably worse results,
	/// if large distances have to be established. A larger fineLevel leads to blurry results.
	/// the number of iterations seems to be more important than the number of GS iterations,
	/// they can even be set to 1. The beta value is for the smoothness, the smaller the better
	/// the results, it seems. If the alpha value is increased the change in u and v is much more
	/// aggressive, leading to faster convergence (usually).
	/// </para>
	/// </remarks>
	/// <author>Martin Eisemann (ME), eisemann@cg.tu-bs.de</author>
	public class OpticalFlow
	{
		#region Public properties
//=================================================================================================

		/// <summary>Gets or sets the number of outer iterations.</summary>
		public int Iterations
		{
			get { return mIterations; }
			set
			{
				mIterations = Math.Max(2, value); // number of Iterations must be at least 2
				mIterations = ((mIterations >> 1) << 1); // ensure it is divisable by 2
			}
		}

		/// <summary>Gets or sets the number of Jacobi iterations.</summary>
		public int JacobiIterations
		{
			get { return mJIterations; }
			set
			{
				mJIterations = value;

				if (mJIterations < 1)
				{
					mJIterations = 1;
					Console.Error.WriteLine("Minimum number of Jacobi iterations is 1.");
				}

				if (mJIterations > 200)
				{
					mJIterations = 200;
					Console.Error.WriteLine("Maximum number of Jacobi iterations is 200.");
				}
			}
		}

		/// <summary>Gets or sets alpha, the gradient constancy weight of the Jacobi iteration.</summary>
		/// <remark>It seems that a large value allows for more rigid transformations.</remark>
		public float Alpha
		{
			get { return mAlpha; }
			set
			{
				mAlpha = value;
				mNeedToUpdateUniforms = true;
			}
		}

		/// <summary>Gets or sets beta, the smoothness weight of the Jacob iteration.</summary>
		public float Beta
		{
			get { return mBeta; }
			set
			{
				mBeta = value;
				mNeedToUpdateUniforms = true;
			}
		}

		/// <summary>Gets or sets the fine level.</summary>
		public int FineLevel
		{
			get { return mFineLevel; }
			set
			{
				if (mInitialised)
					throw new NotSupportedException("Cannot change levels after initialisation.");

				if (value < 0)
					throw new Exception("Cannot set to less than zero.");

				if (value > mCoarseLevel - mLevelOffset)
					throw new Exception(String.Format(
						"Cannot set to more than the coarse level - offset ({0}-{1}).", mCoarseLevel, mLevelOffset));

				mFineLevel = value;
			}
		}

		/// <summary>Gets or sets the coarse level.</summary>
		public int CoarseLevel
		{
			get { return mCoarseLevel; }
			set
			{
				if (mInitialised)
					throw new NotSupportedException("Cannot change levels after initialisation.");

				if (value < mFineLevel + mLevelOffset)
					throw new Exception(String.Format(
						"Cannot set to less than the fine level + offset ({0}+{1}).", mFineLevel, mLevelOffset));

				mCoarseLevel = value;
			}
		}

		/// <summary>Gets or sets the level offset.</summary>
		public int LevelOffset
		{
			get { return mLevelOffset; }
			set
			{
				if (mInitialised)
					throw new NotSupportedException("Cannot change levels after initialisation.");

				if (value < 0)
					throw new Exception("Cannot set to less than zero.");

				if (value > mCoarseLevel - mFineLevel)
					throw new Exception(String.Format(
						"Cannot set to more than than the coarse level - fine level ({0}-{1}).", mCoarseLevel, mFineLevel));

				//mLevelOffset = fineLevel - mFineLevel;
				mLevelOffset = value;
			}
		}

//=================================================================================================
		#endregion

		#region Public methods
//=================================================================================================

		/// <summary>Initialises a new instance of the <see cref="OpticalFlow"/> class.</summary>
		public OpticalFlow()
		{
		}

		/// <summary>Initialises the textures, shaders and FBOs.</summary>
		/// <param name="width">The width of frames.</param>
		/// <param name="height">The height of frames.</param>
		public void Init(int width, int height)
		{
			mNumLevels = mCoarseLevel - mFineLevel + 1;

			mDxy = new float[2, mCoarseLevel + 1];

			mMipmapWidth  = new int[mCoarseLevel + 1];
			mMipmapHeight = new int[mCoarseLevel + 1];

			mGaussTexID = 0;

			mReduceTexID           = new int[mCoarseLevel + 1]; // coarse level+1, because the reduction shader reduces NxN to N/2xN/2
			mReduceHorizontalTexID = new int[mCoarseLevel + 1]; // the coarsest one is not needed, but added for simplicity of code
			mReduceFBO             = new int[mCoarseLevel + 1];

			mDerivativeTexID = new int[5, mNumLevels];
			mDerivativeFBO   = new int[2, mNumLevels];

			mDerivativeProgram[0] = 0;
			mDerivativeProgram[1] = 0;

			mDerivativeUniLocDxy[0] = 0;
			mDerivativeUniLocDxy[1] = 0;
			mDerivativeUniLocDyx[0] = 0;
			mDerivativeUniLocDyx[1] = 0;

			mTensorTexID = new int[4, mNumLevels];
			mTensorFBO   = new int[mNumLevels];

			mPhi3TexID   = new int[mNumLevels];
			mPhi3FBO     = new int[mNumLevels];

			mJacobiFBO   = new int[mNumLevels];

			mUpdateUVFBO = new int[mNumLevels];

			mScaleFBO    = new int[mNumLevels - 1];

			mUVattachpoints  = new DrawBufferMode[2, mNumLevels];
			mDUVattachpoints = new DrawBufferMode[2, mNumLevels];
			for (int j = 0; j < mNumLevels; j++)
			{
				mUVattachpoints [0, j] = DrawBufferMode.ColorAttachment0;
				mUVattachpoints [1, j] = DrawBufferMode.ColorAttachment1;
				mDUVattachpoints[0, j] = DrawBufferMode.ColorAttachment0;
				mDUVattachpoints[1, j] = DrawBufferMode.ColorAttachment1;
			}

			mUVwrite  = Enumerable.Repeat(1, mNumLevels).ToArray();
			mDUVwrite = Enumerable.Repeat(1, mNumLevels).ToArray();

			mUVTexID  = new int[2, mNumLevels];
			mDUVTexID = new int[2, mNumLevels];


			int maxDrawBuffers;
			GL.GetInteger(GetPName.MaxDrawBuffers, out maxDrawBuffers);
			if (maxDrawBuffers < 4)
				throw new Exception("You have " + maxDrawBuffers + " draw buffers available. You need 4.");

			int maxColorAttachments;
			GL.GetInteger(GetPName.MaxColorAttachments, out maxColorAttachments);
			if (maxColorAttachments < 4)
				throw new Exception("Your max. number of color attachments is " + maxColorAttachments + ". You need 4.");

			mWidth  = width;
			mHeight = height;

			for (int i = 0; i <= mCoarseLevel; i++)
			{
				int w = mWidth  >> i;
				int h = mHeight >> i;

				mMipmapWidth[i]  = w;
				mMipmapHeight[i] = h;

				mDxy[0, i] = 1.0f / (float)(w);
				mDxy[1, i] = 1.0f / (float)(h);
			}

			InitTextures();
			InitFBOs();
			InitShaders();

			mInitialised = true;
		}

		/// <summary>Computes the optical flow between the source and target images.</summary>
		/// <param name="source">The source texture.</param>
		/// <param name="target">The target texture.</param>
		/// <param name="result">The result texture, where the optical flow will be rendered into.</param>
		/// <returns>The texture containing the optical flow.</returns>
		public int Compute(int source, int target, int result)
		{
			GL.Enable(EnableCap.Texture2D);
			GL.Disable(EnableCap.DepthTest);
			GL.ClearColor(0, 0, 0, 0);

			if (mNeedToUpdateUniforms)
				UpdateUniforms();

			// swap textures to render into given texture
			int originalResult = mResultUVTexID[0];
			mResultUVTexID[0] = result;
			GL.DeleteFramebuffers(1, ref mScaleToResultFBO);
			mScaleToResultFBO = CreateFBO(mResultUVTexID);
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, 0);

			// preprocessing stage //
			PreProcess(source, target);

			// iteration stage //
			mResultDrawBuffer = DrawBufferMode.ColorAttachment0;
			Iterate(mIterations);

			GL.ClearColor(mClearColor);
			GL.UseProgram(0);

			// restore texture
			mResultUVTexID[0] = originalResult;

			//return CycleTextures(); // for debugging
			return result; // optical flow result
		}

		/// <summary>Computes the optical flow and uses it for morphing between two images.</summary>
		/// <remarks>Not tested, i.e. not compared to original implementation.</remarks>
		/// <param name="morphValue">The morphing value for inbetweening.</param>
		/// <param name="inputTextures">The two input textures.</param>
		public void Render(float morphValue, int[] inputTextures)
		{
			GL.Enable(EnableCap.Texture2D);
			GL.Disable(EnableCap.DepthTest);

			GL.ClearColor(mClearColor);

			// preprocessing stage //
			PreProcess(inputTextures[0], inputTextures[1]);

			// iteration stage //
			mResultDrawBuffer = DrawBufferMode.ColorAttachment0;
			Iterate(mIterations);

			// morph images //
			// 1.0 - morphValue, because the morph is defined in time, not weight
			ApplyOpticalFlow(1.0f - morphValue, 0, inputTextures[0], inputTextures[1], mResultUVTexID[0]);

			GL.Enable(EnableCap.DepthTest);
			GL.ClearColor(mClearColor);
		}

		/// <summary>Computes the optical flow and uses it for morphing between three images.</summary>
		/// <remarks>Not tested, i.e. not compared to original implementation.</remarks>
		/// <param name="morphValue">The morph values for inbetweening.</param>
		/// <param name="inputTextures">The three input textures.</param>
		public void Render(float[] morphValue, int[] inputTextures)
		{
			GL.Enable(EnableCap.Texture2D);
			GL.Disable(EnableCap.DepthTest);
			GL.ClearColor(0, 0, 0, 0);

			// I0 <-> I1 //
			mResultDrawBuffer = DrawBufferMode.ColorAttachment0;
			PreProcess(inputTextures[0], inputTextures[1]);
			Iterate(mIterations);

			// I0 <-> I2 //
			mResultDrawBuffer = DrawBufferMode.ColorAttachment1;
			PreProcess(inputTextures[0], inputTextures[2]);
			Iterate(mIterations);

			// I1 <-> I2 //
			mResultDrawBuffer = DrawBufferMode.ColorAttachment2;
			PreProcess(inputTextures[1], inputTextures[2]);
			Iterate(mIterations);

			// morph images //
			ApplyOpticalFlow3(morphValue, 0, inputTextures, mResultUVTexID);

			GL.Enable(EnableCap.DepthTest);
			GL.ClearColor(mClearColor);
		}

		/// <summary>Computes the optical flow and uses it for morphing between three images.</summary>
		/// <param name="silhouetteTexIDs">The silhouette texture IDs.</param>
		/// <param name="weightTexture">The weight texture.</param>
		/// <param name="inputTextures">The three input textures.</param>
		/// <param name="outputFBO">The output FBO.</param>
		/// <remarks>Not tested, i.e. not compared to original implementation. Silhouettes are ignored in the rest of the code.</remarks>
		public void Render(int[] silhouetteTexIDs, int weightTexture, int[] inputTextures, int outputFBO)
		{
			mSilhouetteTexID = silhouetteTexIDs;

			GL.Disable(EnableCap.DepthTest);
			GL.ClearColor(0.0f, 0.0f, 0.0f, 0.0f);

			// I0 <-> I1 //
			mResultDrawBuffer = DrawBufferMode.ColorAttachment0;
			PreProcess(inputTextures[0], inputTextures[1]);
			Iterate(mIterations);

			// I0 <-> I2 //
			mResultDrawBuffer = DrawBufferMode.ColorAttachment1;
			PreProcess(inputTextures[0], inputTextures[2]);
			Iterate(mIterations);

			// I1 <-> I2 //
			mResultDrawBuffer = DrawBufferMode.ColorAttachment2;
			PreProcess(inputTextures[1], inputTextures[2]);
			Iterate(mIterations);

			// morph images //
			ApplyOpticalFlowWeightMap(weightTexture, outputFBO, inputTextures, mResultUVTexID);

			GL.Enable(EnableCap.DepthTest);
			GL.ClearColor(mClearColor);
		}

		/// <summary>Applies the optical flow to the source images.</summary>
		/// <remarks>Not tested, i.e. not compared to original implementation.</remarks>
		/// <param name="value">The value [0..1]: 0 is texSource0, 1 texSource1.</param>
		/// <param name="outputFBO">The output FBO.</param>
		/// <param name="texSource0">The first source texture.</param>
		/// <param name="texSource1">The second source texture.</param>
		/// <param name="opticalFlowTexture">The texture containing the optical flow texture.</param>
		public void ApplyOpticalFlow(float value, int outputFBO, int texSource0, int texSource1, int opticalFlowTexture)
		{
			// activate program and framebuffer object //
			GL.UseProgram(mMorphProgram);
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, outputFBO);
			GL.Disable(EnableCap.DepthTest);

			// bind textures and set uniforms //
			BindTextures(new int[] { texSource0, texSource1, opticalFlowTexture });

			GL.Uniform4(mMorphUniLocValue, value, value, 1.0f - value, 1.0f - value);

			DrawScreenQuad(mMipmapWidth[0], mMipmapHeight[0]);

			GL.ActiveTexture(TextureUnit.Texture0);
		}

		/// <summary>Applies the optical flow to the source images.</summary>
		/// <remarks>Not tested, i.e. not compared to original implementation.</remarks>
		/// <param name="value">The weight for every image, should sum up to 1.</param>
		/// <param name="outputFBO">The FBO where the result is rendered to.</param>
		/// <param name="texSource">The input images.</param>
		/// <param name="opticalFlowTexture">The The flowfields used for warping.</param>
		public void ApplyOpticalFlow3(float[] value, int outputFBO, int[] texSource, int[] opticalFlowTexture)
		{
			// activate program and framebuffer object //
			GL.UseProgram(mMorph3Program);
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, outputFBO);

			// bind textures and set uniforms //
			BindTextures(new int[] {
				texSource[0], texSource[1], texSource[2],
				opticalFlowTexture[0], opticalFlowTexture[1], opticalFlowTexture[2]
			});

			GL.Uniform3(mMorph3UniLocWeights, value[0], value[1], value[2]);

			DrawScreenQuad(mMipmapWidth[0], mMipmapHeight[0]);

			GL.ActiveTexture(TextureUnit.Texture0);
		}

		/// <summary>Applies the optical flow to the source images.</summary>
		/// <remarks>Not tested, i.e. not compared to original implementation.</remarks>
		/// <param name="weightTex">texture with per-pixel weights(?).</param>
		/// <param name="outputFBO">The FBO where the result is rendered to.</param>
		/// <param name="texSource">The input images.</param>
		/// <param name="opticalFlowTexture">The The flowfields used for warping.</param>
		public void ApplyOpticalFlowWeightMap(int weightTex, int outputFBO, int[] texSource, int[] opticalFlowTexture)
		{
			// activate program and framebuffer object //
			GL.UseProgram(mMorphWeightMapProgram);
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, outputFBO);

			// bind textures and set uniforms //
			BindTextures(new int[] {
				texSource[0], texSource[1], texSource[2],
				opticalFlowTexture[0], opticalFlowTexture[1], opticalFlowTexture[2],
				weightTex
			});

			DrawScreenQuad(mMipmapWidth[0], mMipmapHeight[0]);

			GL.ActiveTexture(TextureUnit.Texture0);
		}

		/// <summary>Returns the optical flow texture.</summary>
		/// <param name="idx">Texture index (0:0-1, 1:0-2, 2:1-1).</param>
		public int GetFlowTexture(int idx)
		{
			if (idx < 0) idx = 0;
			if (idx > 2) idx = 2;
			return mResultUVTexID[idx];
		}

//=================================================================================================
		#endregion
		
		#region Protected methods
//=================================================================================================
		
		#region Texture & FBO initialisation
//-------------------------------------------------------------------------------------------------

		/// <summary>Initialises the textures.</summary>
		protected void InitTextures()
		{
			for (int j = 0; j < mNumLevels; j++)
			{
				int level = mFineLevel + j;

				for (int i = 0; i < 2; i++)
				{
					mUVTexID[i, j] = CreateTextureDataf(IntPtr.Zero, mMipmapWidth[level], mMipmapHeight[level], mTarget, mIFormat, mFormat, mType, false,
						TextureMagFilter.Linear, TextureMinFilter.Linear, TextureWrapMode.Clamp, TextureWrapMode.Clamp, false);

					mDUVTexID[i, j] = CreateTextureDataf(IntPtr.Zero, mMipmapWidth[level], mMipmapHeight[level], mTarget, mIFormat, mFormat, mType, false,
						TextureMagFilter.Linear, TextureMinFilter.Linear, TextureWrapMode.Clamp, TextureWrapMode.Clamp, false);
				}
			}
		}

		/// <summary>Initialises the framebuffer objects.</summary>
		protected void InitFBOs()
		{
			// initialize the FBOs for every pass //
			InitGrayFBOs();
			InitReductionFBOs(); // needs to be initialized before the gaussian
			InitGaussianFBOs();
			InitDerivativesFBOs();
			InitTensorFBOs();
			InitPhi3FBOs();
			InitJacobiFBOs();
			InitUpdateUVFBOs();
			InitScaleFBOs();
			InitScaleToResultFBOs();
			InitResetUVFBOs();

			// restore window frame buffer //
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, 0);
		}

		/// <summary>Initialises the framebuffer objects for the rgb2gray pass.</summary>
		protected void InitGrayFBOs()
		{
			// convert to gray //
			mGrayTexID = CreateFBOTexture(mWidth, mHeight);
			mGrayFBO   = CreateFBO(mGrayTexID);

			GL.BindFramebuffer(FramebufferTarget.Framebuffer, 0);
		}

		/// <summary>Initialises the framebuffer objects for the reduction.</summary>
		protected void InitReductionFBOs()
		{
			// reduction (has to be initialised before Gaussian) //
			for (int i = 0; i <= mCoarseLevel; i++)
			{
				mReduceTexID[i]           = CreateFBOTexture(mMipmapWidth[i], mMipmapHeight[i]);
				mReduceHorizontalTexID[i] = CreateFBOTexture(mMipmapWidth[i], mMipmapHeight[i]);
				mReduceFBO[i]             = CreateFBO(new int[] { mReduceHorizontalTexID[i], mReduceTexID[i] });
			}

			GL.BindFramebuffer(FramebufferTarget.Framebuffer, 0);
		}

		/// <summary>Initialises the framebuffer objects for the Gaussian presmoothing.</summary>
		protected void InitGaussianFBOs()
		{
			// Gaussian smoothing //
			mGaussTexID = CreateFBOTexture(mWidth, mHeight);
			mGaussFBO   = CreateFBO(new int[] { mGaussTexID, mReduceTexID[0] });

			GL.BindFramebuffer(FramebufferTarget.Framebuffer, 0);
		}

		/// <summary>Initialises the framebuffer objects for the derivation pass.</summary>
		protected void InitDerivativesFBOs()
		{
			// derivatives //
			for (int i = 0; i < mNumLevels; i++)
			{
				int level = mFineLevel + i;

				// textures
				mDerivativeTexID[(int)Derivative.DX,  i] = CreateFBOTexture(mMipmapWidth[level], mMipmapHeight[level]);
				mDerivativeTexID[(int)Derivative.DY,  i] = CreateFBOTexture(mMipmapWidth[level], mMipmapHeight[level]);
				mDerivativeTexID[(int)Derivative.DXX, i] = CreateFBOTexture(mMipmapWidth[level], mMipmapHeight[level]);
				mDerivativeTexID[(int)Derivative.DXY, i] = CreateFBOTexture(mMipmapWidth[level], mMipmapHeight[level]);
				mDerivativeTexID[(int)Derivative.DYY, i] = CreateFBOTexture(mMipmapWidth[level], mMipmapHeight[level]);

				// first fbo
				mDerivativeFBO[0, i] = CreateFBO(new int[] {
					mDerivativeTexID[(int)Derivative.DX, i],
					mDerivativeTexID[(int)Derivative.DY, i]
				});

				// second fbo
				mDerivativeFBO[1, i] = CreateFBO(new int[] {
					mDerivativeTexID[(int)Derivative.DXX, i],
					mDerivativeTexID[(int)Derivative.DXY, i],
					mDerivativeTexID[(int)Derivative.DYY, i]
				});
			}

			GL.BindFramebuffer(FramebufferTarget.Framebuffer, 0);
		}

		/// <summary>Initialises the framebuffer objects for the warping pass.</summary>
		protected void InitTensorFBOs()
		{
			for (int i = 0; i < mNumLevels; i++)
			{
				int level = mFineLevel + i;
				mTensorTexID[0, i] = CreateFBOTexture(mMipmapWidth[level], mMipmapHeight[level]);
				mTensorTexID[1, i] = CreateFBOTexture(mMipmapWidth[level], mMipmapHeight[level]);
				mTensorTexID[2, i] = CreateFBOTexture(mMipmapWidth[level], mMipmapHeight[level]);
				mTensorTexID[3, i] = CreateFBOTexture(mMipmapWidth[level], mMipmapHeight[level]);

				mTensorFBO[i] = CreateFBO(new int[] {
					mTensorTexID[0,i],
					mTensorTexID[1,i],
					mTensorTexID[2,i],
					mTensorTexID[3,i]
				});
			}

			GL.BindFramebuffer(FramebufferTarget.Framebuffer, 0);
		}

		///<summary>Initialises the frambuffer objects for the calculation of phi3.</summary>
		protected void InitPhi3FBOs()
		{
			for (int i = 0; i < mNumLevels; i++)
			{
				int level = mFineLevel + i;
				mPhi3TexID[i] = CreateFBOTexture(mMipmapWidth[level], mMipmapHeight[level]);
				mPhi3FBO[i]   = CreateFBO(mPhi3TexID[i]);
			}

			GL.BindFramebuffer(FramebufferTarget.Framebuffer, 0);
		}

		///<summary>Initialises the frambuffer objects for the Jacobi solver.</summary>
		protected void InitJacobiFBOs()
		{
			for (int i = 0; i < mNumLevels; i++)
			{
				int readDUV  = GetReadTex(mDUVwrite[i]);
				int writeDUV = mDUVwrite[i];

				mJacobiFBO[i] = CreateFBO(new int[] {
					mDUVTexID[readDUV,  i],
					mDUVTexID[writeDUV, i]
				});
			}

			GL.BindFramebuffer(FramebufferTarget.Framebuffer, 0);
		}

		///<summary>Initialises the framebuffer objects for the update of UV (UV = UV + DUV).</summary>
		protected void InitUpdateUVFBOs()
		{
			for (int i = 0; i < mNumLevels; i++)
			{
				int writeUV  = mUVwrite[i];
				int writeDUV = mDUVwrite[i];
				int readUV   = GetReadTex(mUVwrite[i]);
				int readDUV  = GetReadTex(mDUVwrite[i]);

				mUpdateUVFBO[i] = CreateFBO(new int[] {
					mUVTexID[ readUV,  i],
					mUVTexID[ writeUV, i],
					mDUVTexID[readDUV, i],
					mDUVTexID[writeDUV,i]
				});

				DrawBuffersEnum[] updateUVDrawbuffers = { DrawBuffersEnum.ColorAttachment1, DrawBuffersEnum.ColorAttachment3 };
				GL.DrawBuffers(2, updateUVDrawbuffers);

				CheckFboStatus();
			}

			GL.BindFramebuffer(FramebufferTarget.Framebuffer, 0);
		}

		///<summary>Initialises the framebuffer objects for the upscale pass.</summary>
		protected void InitScaleFBOs()
		{
			for (int i = 0; i < mNumLevels - 1; i++)
			{
				int readUV  = GetReadTex(mUVwrite[i]);
				int writeUV = mUVwrite[i];

				mScaleFBO[i] = CreateFBO(new int[] {
					mUVTexID[readUV, i],
					mUVTexID[writeUV, i]
				});
			}

			GL.BindFramebuffer(FramebufferTarget.Framebuffer, 0);
		}

		///<summary>Initialises the framebuffer objects for the final scale.</summary>
		protected void InitScaleToResultFBOs()
		{
			mResultUVTexID[0] = CreateFBOTexture(mMipmapWidth[0], mMipmapHeight[0]);
			mResultUVTexID[1] = CreateFBOTexture(mMipmapWidth[0], mMipmapHeight[0]);
			mResultUVTexID[2] = CreateFBOTexture(mMipmapWidth[0], mMipmapHeight[0]);
			mScaleToResultFBO = CreateFBO(mResultUVTexID);

			GL.BindFramebuffer(FramebufferTarget.Framebuffer, 0);
		}

		///<summary>Initialise the framebuffer for clearing the coarsest UV texture in the end.</summary>
		protected void InitResetUVFBOs()
		{
			int read  = GetReadTex(mUVwrite[mUVwrite.Length - 1]);
			int write = mUVwrite[mUVwrite.Length - 1];
			int level = mNumLevels - 1;

			mResetUVFBO = CreateFBO(new int[] {
				mUVTexID[read,  level],
				mUVTexID[write, level]
			});

			GL.BindFramebuffer(FramebufferTarget.Framebuffer, 0);
		}

//-------------------------------------------------------------------------------------------------
		#endregion

		#region Shader initialisation
//-------------------------------------------------------------------------------------------------

		///<summary>Initialises the shaders.</summary>
		protected void InitShaders()
		{
			// rgb2gray shader program //
			InitGrayShader();

			// Gaussian smoothing //
			InitGaussShader();

			// reduction //
			InitReduceShader();

			// derivatives //
			InitDerivativesShader();

			// warping shader program //
			InitTensorShader();

			// optimization shader programs //
			InitPhi3Shader();
			InitJacobiShader();
			InitUpdateUVShader();
			InitScaleShader();

			// final scale //
			InitScaleToResultShader();

			// morph shader program //
			InitMorphShader();
			InitMorph3Shader();
			InitMorphWeightMapShader();

			UpdateUniforms();

			// switch back to standard pipeline //
			GL.UseProgram(0);
		}

		///<summary>Initialises the rgb2gray shader.</summary>
		protected void InitGrayShader()
		{
			mGrayProgram = CreateShaderProgram(new string[0], new string[] {
				"Shaders.ImageProcessing.frag",
				"Shaders.Rgb2gray.frag"
			});

			GL.UseProgram(mGrayProgram);

			GL.Uniform1(GL.GetUniformLocation(mGrayProgram, "tex0"), 0);
			GL.Uniform1(GL.GetUniformLocation(mGrayProgram, "tex1"), 1);
			GL.Uniform1(GL.GetUniformLocation(mGrayProgram, "tex2"), 2);
		}

		///<summary>Initialises the Gaussian smoothing shader.</summary>
		protected void InitGaussShader()
		{
			float[] mask = new float[] { 0.1201f, 0.2339f, 0.292f, 0.2339f, 0.1201f };

			mGaussProgram = CreateShaderProgram(new string[0], new string[] {
				"Shaders.Linear1D5Filter.frag",
				"Shaders.ImageProcessing.frag"
			});

			GL.UseProgram(mGaussProgram);
			mGaussUniLocDxy = GL.GetUniformLocation(mGaussProgram, "dxy");

			GL.Uniform1(GL.GetUniformLocation(mGaussProgram, "source"), 0);
			GL.Uniform1(GL.GetUniformLocation(mGaussProgram, "mask"), 5, mask);
		}

		///<summary>Initialises the reduction shader.</summary>
		protected void InitReduceShader()
		{
			mReduceProgram = CreateShaderProgram(new string[0], new string[] {
				"Shaders.Reduce.frag",
				"Shaders.ImageProcessing.frag"
			});

			GL.UseProgram(mReduceProgram);
			mReduceUniLocDxy = GL.GetUniformLocation(mReduceProgram, "dxy");
			GL.Uniform1(GL.GetUniformLocation(mReduceProgram, "source"), 0);
		}

		///<summary>Initialises the shader for the first and second derivative.</summary>
		protected void InitDerivativesShader()
		{
			float[] mask = new float[] { 0.0833f, -0.6667f, 0.0f, 0.6667f, -0.0833f };

			// first derivative shader //
			mDerivativeProgram[0] = CreateShaderProgram(new string[0], new string[] {
				"Shaders.Derivatives1D5.frag",
				"Shaders.ImageProcessing.frag"
			});

			GL.UseProgram(mDerivativeProgram[0]);
			mDerivativeUniLocDxy[0] = GL.GetUniformLocation(mDerivativeProgram[0], "dxy");
			mDerivativeUniLocDyx[0] = GL.GetUniformLocation(mDerivativeProgram[0], "dyx");
			GL.Uniform1(GL.GetUniformLocation(mDerivativeProgram[0], "source"), 0);
			GL.Uniform1(GL.GetUniformLocation(mDerivativeProgram[0], "mask"), 5, mask);

			// second derivative shader //
			mDerivativeProgram[1] = CreateShaderProgram(new string[0], new string[] {
				"Shaders.SecondDerivatives1D5.frag",
				"Shaders.ImageProcessing.frag"
			});

			GL.UseProgram(mDerivativeProgram[1]);
			mDerivativeUniLocDxy[1] = GL.GetUniformLocation(mDerivativeProgram[1], "dxy");
			mDerivativeUniLocDyx[1] = GL.GetUniformLocation(mDerivativeProgram[1], "dyx");
			GL.Uniform1(GL.GetUniformLocation(mDerivativeProgram[1], "sourcex"), 0);
			GL.Uniform1(GL.GetUniformLocation(mDerivativeProgram[1], "sourcey"), 1);
			GL.Uniform1(GL.GetUniformLocation(mDerivativeProgram[1], "mask"), 5, mask);
		}

		///<summary>Initialises the warping shader.</summary>
		protected void InitTensorShader()
		{
			mTensorProgram = CreateShaderProgram(new string[0], new string[] {
				"Shaders.CalculateTensor.frag"
			});

			GL.UseProgram(mTensorProgram);
			mTensorUniLocDxy = GL.GetUniformLocation(mTensorProgram, "dxy");
			GL.Uniform1(GL.GetUniformLocation(mTensorProgram, "texSource"), 0);
			GL.Uniform1(GL.GetUniformLocation(mTensorProgram, "texDX"), 1);
			GL.Uniform1(GL.GetUniformLocation(mTensorProgram, "texDY"), 2);
			GL.Uniform1(GL.GetUniformLocation(mTensorProgram, "texDXX"), 3);
			GL.Uniform1(GL.GetUniformLocation(mTensorProgram, "texDXY"), 4);
			GL.Uniform1(GL.GetUniformLocation(mTensorProgram, "texDYY"), 5);
			GL.Uniform1(GL.GetUniformLocation(mTensorProgram, "texUVField"), 6);
		}

		///<summary>Initialises the optimization shader.</summary>
		protected void InitPhi3Shader()
		{
			mPhi3Program = CreateShaderProgram(new string[0], new string[] {
				"Shaders.Phi3.frag"
			});

			GL.UseProgram(mPhi3Program);
			mPhi3UniLocDxy = GL.GetUniformLocation(mPhi3Program, "dxy");
			GL.Uniform1(GL.GetUniformLocation(mPhi3Program, "texUVField"), 0);
			GL.Uniform1(GL.GetUniformLocation(mPhi3Program, "texDUVField"), 1);
			//GL.Uniform1(GL.GetUniformLocation(mPhi3Program, "silhouette"), 2);
		}

		///<summary>Initialises the Jacobi iteration shader.</summary>
		protected void InitJacobiShader()
		{
			mJacobiProgram = CreateShaderProgram(new string[0], new string[] {
				"Shaders.JacobiAndUpdateUV.frag"
			});

			GL.UseProgram(mJacobiProgram);
			mJacobiUniLocDxy = GL.GetUniformLocation(mJacobiProgram, "dxy");

			GL.Uniform2(GL.GetUniformLocation(mJacobiProgram, "alphav2"), mAlpha, mAlpha);
			GL.Uniform4(GL.GetUniformLocation(mJacobiProgram, "alphav4"), mAlpha, mAlpha, mAlpha, mAlpha);
			GL.Uniform4(GL.GetUniformLocation(mJacobiProgram, "betav4"), mBeta, mBeta, mBeta, mBeta);

			GL.Uniform1(GL.GetUniformLocation(mJacobiProgram, "texTensor0"), 0);
			GL.Uniform1(GL.GetUniformLocation(mJacobiProgram, "texTensor1"), 1);
			GL.Uniform1(GL.GetUniformLocation(mJacobiProgram, "texTensor2"), 2);
			GL.Uniform1(GL.GetUniformLocation(mJacobiProgram, "texTensor3"), 3);

			GL.Uniform1(GL.GetUniformLocation(mJacobiProgram, "texUVField"), 4);

			GL.Uniform1(GL.GetUniformLocation(mJacobiProgram, "texPhi3"), 6);

			// ignore silhouette (CR 2010-08-20)
			//GL.Uniform1(GL.GetUniformLocation(mJacobiProgram, "silhouette"), 7);
		}

		///<summary>Initialises the shader for the update of UV.</summary>
		protected void InitUpdateUVShader()
		{
			mUpdateUVProgram = CreateShaderProgram(new string[0], new string[] {
				"Shaders.UpdateUV.frag"
			});

			GL.UseProgram(mUpdateUVProgram);
			GL.Uniform1(GL.GetUniformLocation(mUpdateUVProgram, "texUVField"), 0);
			GL.Uniform1(GL.GetUniformLocation(mUpdateUVProgram, "texDUVField"), 1);
		}

		///<summary>Initialises the shader for the scaling of UV to the next level.</summary>
		protected void InitScaleShader()
		{
			mScaleProgram = CreateShaderProgram(new string[0], new string[] {
				"Shaders.Scale.frag"
			});

			GL.UseProgram(mScaleProgram);
			GL.Uniform1(GL.GetUniformLocation(mScaleProgram, "texUVField"), 0);
		}

		///<summary>Initialises the shader for the final scaling of the UV to the next level.</summary>
		protected void InitScaleToResultShader()
		{
			mScaleToResultProgram = CreateShaderProgram(new string[0], new string[] {
				"Shaders.ScaleToResult.frag"
			});

			GL.UseProgram(mScaleToResultProgram);
			GL.Uniform1(GL.GetUniformLocation(mScaleToResultProgram, "texUVField"), 0);
			mScaleToResultUniLocScale = GL.GetUniformLocation(mScaleToResultProgram, "scale");
		}

		///<summary>Initialises the morphing shader.</summary>
		protected void InitMorphShader()
		{
			mMorphProgram = CreateShaderProgram(new string[0], new string[] {
				"Shaders.Morph.frag"
			});

			GL.UseProgram(mMorphProgram);
			mMorphUniLocValue = GL.GetUniformLocation(mMorphProgram, "value");
			GL.Uniform1(GL.GetUniformLocation(mMorphProgram, "source"), 0);
			GL.Uniform1(GL.GetUniformLocation(mMorphProgram, "target"), 1);
			GL.Uniform1(GL.GetUniformLocation(mMorphProgram, "opticalflow"), 2);
			GL.Uniform4(GL.GetUniformLocation(mMorphProgram, "dxy"), mDxy[0, 0], -mDxy[1, 0], mDxy[0, 0], -mDxy[1, 0]);
		}

		///<summary>Initialises the morphing shader for morphing between 3 images.</summary>
		protected void InitMorph3Shader()
		{
			mMorph3Program = CreateShaderProgram(new string[0], new string[] {
				"Shaders.Morph3.frag"
			});

			GL.UseProgram(mMorph3Program);
			mMorph3UniLocWeights = GL.GetUniformLocation(mMorph3Program, "weights");
			GL.Uniform1(GL.GetUniformLocation(mMorph3Program, "source0"), 0);
			GL.Uniform1(GL.GetUniformLocation(mMorph3Program, "source1"), 1);
			GL.Uniform1(GL.GetUniformLocation(mMorph3Program, "source2"), 2);
			GL.Uniform1(GL.GetUniformLocation(mMorph3Program, "opticalflow0"), 3);
			GL.Uniform1(GL.GetUniformLocation(mMorph3Program, "opticalflow1"), 4);
			GL.Uniform1(GL.GetUniformLocation(mMorph3Program, "opticalflow2"), 5);
			GL.Uniform2(GL.GetUniformLocation(mMorph3Program, "dxy"), mDxy[0, 0], -mDxy[1, 0]); // - because y-axis is mirrored in matlab
		}

		///<summary>Initialises the morphing shader between 3 images by applying a weight map.</summary>
		protected void InitMorphWeightMapShader()
		{
			mMorphWeightMapProgram = CreateShaderProgram(new string[0], new string[] {
				"Shaders.MorphWeightMap.frag"
			});

			GL.UseProgram(mMorphWeightMapProgram);
			GL.Uniform1(GL.GetUniformLocation(mMorphWeightMapProgram, "source0"), 0);
			GL.Uniform1(GL.GetUniformLocation(mMorphWeightMapProgram, "source1"), 1);
			GL.Uniform1(GL.GetUniformLocation(mMorphWeightMapProgram, "source2"), 2);
			GL.Uniform1(GL.GetUniformLocation(mMorphWeightMapProgram, "opticalflow0"), 3);
			GL.Uniform1(GL.GetUniformLocation(mMorphWeightMapProgram, "opticalflow1"), 4);
			GL.Uniform1(GL.GetUniformLocation(mMorphWeightMapProgram, "opticalflow2"), 5);
			GL.Uniform1(GL.GetUniformLocation(mMorphWeightMapProgram, "weightMap"), 6);
			GL.Uniform2(GL.GetUniformLocation(mMorphWeightMapProgram, "dxy"), mDxy[0, 0], -mDxy[1, 0]); // - because y-axis is mirrored in matlab
		}

//-------------------------------------------------------------------------------------------------
		#endregion

		#region Computation methods
//-------------------------------------------------------------------------------------------------

		/// <summary>Updates the uniforms in the shaders.</summary>
		protected void UpdateUniforms()
		{
			GL.UseProgram(mJacobiProgram);
			GL.Uniform2(GL.GetUniformLocation(mJacobiProgram, "alphav2"), mAlpha, mAlpha);
			GL.Uniform4(GL.GetUniformLocation(mJacobiProgram, "alphav4"), mAlpha, mAlpha, mAlpha, mAlpha);
			GL.Uniform4(GL.GetUniformLocation(mJacobiProgram, "betav4"), mBeta, mBeta, mBeta, mBeta);
			GL.UseProgram(0);

			mNeedToUpdateUniforms = false;
		}

		/// <summary>Preprocessing of the data. Information needed for the iteration process is established.</summary>
		/// <param name="inputTexture0">Input texture 0.</param>
		/// <param name="inputTexture1">Input texture 1.</param>
		protected void PreProcess(int inputTexture0, int inputTexture1)
		{
			// create gray images //
			CreateGrayTexture(inputTexture0, inputTexture1);

			// presmooth image //
			PreSmoothStep(); //maybe not really necessary

			// create image pyramid //
			ReductionStep();

			// calculate the first and second derivatives for the image pyramid //
			CalculateDerivatives();
		}

		/// <summary>The optical flow field is established in this stage.</summary>
		/// <param name="iterations">The number of outer iterations.</param>
		protected void Iterate(int iterations)
		{
			// outer iteration //
			for (int i = mNumLevels - 1; i >= 0 + mLevelOffset; i--)
			{
				// start with coarse grid
				for (int j = 0; j < iterations; j++)
				{
					CalculateTensor(i);

					// for(int k=0; k < mJIterations; k++){
					CalculatePhi3(i);

					// calculateJacobiSolver(i);
					CalculateJacobiSolverAndUpdateUV(i);

					// }
					// updateUV(i); //also resets duv to 0
				}

				// scale results to next finer level. from i to i-1 //
				if (i != 0) Scale(i);
			}

			// scale the UV flow field to the finest level and //
			int scaleFactor = 1 << (mFineLevel + mLevelOffset);
			ScaleToResult(scaleFactor);

			// reset coarse UV to 0 (the others are implicitly reset in the scale function) //
			int readUV = mUVwrite[mNumLevels - 1] ^ 1;
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, mResetUVFBO);
			GL.DrawBuffer(mUVattachpoints[readUV, mNumLevels - 1]);
			GL.Clear(ClearBufferMask.ColorBufferBit);
		}

		/// <summary>Converts the input textures to gray values and combines them in one texture.</summary>
		/// <param name="inputTexture0">Input texture 0.</param>
		/// <param name="inputTexture1">Input texture 1.</param>
		protected void CreateGrayTexture(int inputTexture0, int inputTexture1)
		{
			// activate program and framebuffer object //
			GL.UseProgram(mGrayProgram);
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, mGrayFBO);

			// bind textures and set uniforms //
			BindTextures(new int[] { inputTexture0, inputTexture1 });

			GL.Uniform1(GL.GetUniformLocation(mGrayProgram, "numInputs"), 2);

			DrawScreenQuad(mMipmapWidth[0], mMipmapHeight[0]);
			GL.ActiveTexture(TextureUnit.Texture0);

		}

		///<summary>Presmoothes the image with a Gaussian 5x5 filter, mu=0, sigma=1.5.</summary>
		protected void PreSmoothStep()
		{
			// activate program and framebuffer object //
			GL.UseProgram(mGaussProgram);
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, mGaussFBO);

			// horizontal pass //
			GL.DrawBuffer(DrawBufferMode.ColorAttachment0);
			GL.ActiveTexture(TextureUnit.Texture0);
			GL.BindTexture(TextureTarget.Texture2D, mGrayTexID);
			GL.Uniform2(mGaussUniLocDxy, mDxy[0, 0], 0.0f);
			DrawScreenQuad(mMipmapWidth[0], mMipmapHeight[0]);

			// vertical pass //
			GL.DrawBuffer(DrawBufferMode.ColorAttachment1);
			GL.BindTexture(TextureTarget.Texture2D, mGaussTexID);
			GL.Uniform2(mGaussUniLocDxy, 0.0f, -mDxy[1, 0]); // - because axis is mirrored in matlab
			DrawScreenQuad(mMipmapWidth[0], mMipmapHeight[0]);
		}

		/// <summary>Filters and reduces resolution of the images.</summary>
		/// <remarks>Currently only simple boxfilter used for reduction.</remarks>
		protected void ReductionStep()
		{
			GL.UseProgram(mReduceProgram);
			GL.ActiveTexture(TextureUnit.Texture0);
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, mReduceFBO[0]);

			for (int level = 0; level < mCoarseLevel; level++)
			{
				// horizontal filter pass //
				GL.DrawBuffer(DrawBufferMode.ColorAttachment0);
				GL.BindTexture(mTarget, mReduceTexID[level]);
				GL.Uniform2(mReduceUniLocDxy, mDxy[0, level], 0.0f);
				DrawScreenQuad(mMipmapWidth[level], mMipmapHeight[level]);

				GL.BindFramebuffer(FramebufferTarget.Framebuffer, mReduceFBO[level + 1]);

				// vertical filter pass and reduction //
				GL.DrawBuffer(DrawBufferMode.ColorAttachment1);
				GL.BindTexture(mTarget, mReduceHorizontalTexID[level]);
				GL.Uniform2(mReduceUniLocDxy, 0.0f, -mDxy[0, level]); // - because axis is mirrored in matlab
				DrawScreenQuad(mMipmapWidth[level + 1], mMipmapHeight[level + 1]);
			}
		}

		///<summary>Calculate the first and second derivatives of the different smoothed gray input images.</summary>
		protected void CalculateDerivatives()
		{
			GL.ActiveTexture(TextureUnit.Texture0);

			// first derivatives //
			GL.UseProgram(mDerivativeProgram[0]);

			// for every resolution
			for (int i = 0; i < mNumLevels; i++)
			{
				int level = mFineLevel + i;
				GL.BindFramebuffer(FramebufferTarget.Framebuffer, mDerivativeFBO[0, i]);
				GL.BindTexture(mTarget, mReduceTexID[level]);

				GL.Uniform2(mDerivativeUniLocDxy[0], mDxy[0, level], 0.0f);
				GL.Uniform2(mDerivativeUniLocDyx[0], 0.0f, -mDxy[1, level]); // - because axis is mirrored in matlab

				DrawScreenQuad(mMipmapWidth[level], mMipmapHeight[level]);
			}

			// second derivatives //
			GL.UseProgram(mDerivativeProgram[1]);

			for (int i = 0; i < mNumLevels; i++)
			{
				int level = mFineLevel + i;

				GL.BindFramebuffer(FramebufferTarget.Framebuffer, mDerivativeFBO[1, i]);

				BindTextures(new int[] {
					mDerivativeTexID[(int)Derivative.DX, i],
					mDerivativeTexID[(int)Derivative.DY, i]
				});

				GL.Uniform2(mDerivativeUniLocDxy[1], mDxy[0, level], 0.0f);
				GL.Uniform2(mDerivativeUniLocDyx[1], 0.0f, -mDxy[1, level]); // - because axis is mirrored in matlab

				DrawScreenQuad(mMipmapWidth[level], mMipmapHeight[level]);
			}

			GL.ActiveTexture(TextureUnit.Texture0);
		}

		/// <summary>Calculates the warped image and derivatives with the current flow field.</summary>
		/// <param name="i">The current "level" (used for acessing the FBOs and textures). Actual level is mFineLevel + i.</param>
		protected void CalculateTensor(int i)
		{
			Debug.Assert(i < mTensorFBO.Length);
			int level = mFineLevel + i; // the "real" level
			int readUV = GetReadTex(mUVwrite[i]);

			// activate program and framebuffer object //
			GL.UseProgram(mTensorProgram);
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, mTensorFBO[i]);

			// bind textures and set uniforms //
			GL.Uniform2(mTensorUniLocDxy, mDxy[0, level], -mDxy[1, level]); // - because axis is mirrored in matlab

			BindTextures(new int[] {
				mReduceTexID[level],
				mDerivativeTexID[(int)Derivative.DX,i],
				mDerivativeTexID[(int)Derivative.DY,i],
				mDerivativeTexID[(int)Derivative.DXX,i],
				mDerivativeTexID[(int)Derivative.DXY,i],
				mDerivativeTexID[(int)Derivative.DYY,i],
				mUVTexID[readUV,i]
			});

			DrawScreenQuad(mMipmapWidth[level], mMipmapHeight[level]);

			GL.ActiveTexture(TextureUnit.Texture0);
		}

		/// <summary>Calculates phi3 for the Jacobi solver.</summary>
		/// <param name="i">The current "level" (used for acessing the FBOs and textures). Actual level is mFineLevel + i.</param>
		protected void CalculatePhi3(int i)
		{
			// phi3 pass //
			int level = mFineLevel + i;
			int readUV = GetReadTex(mUVwrite[i]);
			int readDUV = GetReadTex(mDUVwrite[i]);

			// activate program and framebuffer object //
			GL.UseProgram(mPhi3Program);
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, mPhi3FBO[i]);

			// bind textures and set uniforms //
			GL.Uniform2(mPhi3UniLocDxy, mDxy[0, level], -mDxy[1, level]); // - because of mirrored axis in matlab

			BindTextures(new int[] {
				mUVTexID[readUV, level],
				mDUVTexID[readDUV, level]
			});

			DrawScreenQuad(mMipmapWidth[level], mMipmapHeight[level]);

			GL.ActiveTexture(TextureUnit.Texture0);
		}

		/// <summary>Calculates one iteration of the Jacobi solver, currently not used.</summary>
		/// <param name="i">The current "level" (used for acessing the FBOs and textures). Actual level is mFineLevel+i.</param>
		/// <remarks>Not tested, i.e. not compared to original implementation.</remarks>
		protected void CalculateJacobiSolver(int i)
		{
			int level = mFineLevel + i;
			int readUV = GetReadTex(mUVwrite[i]);

			// activate program and framebuffer object //
			GL.UseProgram(mJacobiProgram);
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, mJacobiFBO[i]);
			GL.DrawBuffer(mDUVattachpoints[mDUVwrite[i], i]);

			// bind textures and set uniforms //
			GL.Uniform2(mJacobiUniLocDxy, mDxy[0, level], -mDxy[1, level]); // - because axis is mirrored in matlab

			GL.ActiveTexture(TextureUnit.Texture0);
			GL.BindTexture(mTarget, mTensorTexID[0, i]);
			GL.ActiveTexture(TextureUnit.Texture1);
			GL.BindTexture(mTarget, mTensorTexID[1, i]);
			GL.ActiveTexture(TextureUnit.Texture2);
			GL.BindTexture(mTarget, mTensorTexID[2, i]);
			GL.ActiveTexture(TextureUnit.Texture3);
			GL.BindTexture(mTarget, mTensorTexID[3, i]);
			GL.ActiveTexture(TextureUnit.Texture4);
			GL.BindTexture(mTarget, mUVTexID[readUV, i]);
			GL.ActiveTexture(TextureUnit.Texture6);
			GL.BindTexture(mTarget, mPhi3TexID[i]);
			GL.ActiveTexture(TextureUnit.Texture7);
			GL.BindTexture(mTarget, mSilhouetteTexID[i]);

			DrawScreenQuad(mMipmapWidth[level], mMipmapHeight[level]);

			GL.ActiveTexture(TextureUnit.Texture0);

			// swap for pingpong
			mDUVwrite[i] ^= 1; // duv was changed
		}

		/// <summary>Adds DUV to UV for the next iteration pass</summary>
		/// <remarks>Not tested, i.e. not compared to original implementation.</remarks>
		/// <param name="i">The current "level". Actual level is mFineLevel+i.</param>
		protected void UpdateUV(int i)
		{
			int level = mFineLevel + i;
			int writeUV = mUVwrite[i];	// In C++ code, int& writeUV = mUVwrite[i];
			int writeDUV = mDUVwrite[i];// In C++ code, int& writeDUV = mDUVwrite[i];
			int readUV = GetReadTex(writeUV);
			int readDUV = GetReadTex(writeDUV);

			DrawBuffersEnum[] db = {
				(DrawBuffersEnum)mUVattachpoints[writeUV,i],
				(DrawBuffersEnum)(mDUVattachpoints[writeDUV,i] + 2) // attached to color attachment 2 and 3
			};

			// activate program and framebuffer object //
			GL.UseProgram(mUpdateUVProgram);
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, mUpdateUVFBO[i]);
			GL.DrawBuffers(2, db);

			BindTextures(new int[] {
				mUVTexID[readUV, i],
				mDUVTexID[readDUV, i]
			});

			DrawScreenQuad(mMipmapWidth[level], mMipmapHeight[level]);

			GL.ActiveTexture(TextureUnit.Texture0);

			mUVwrite[i] ^= 1; // uv was changed
			mDUVwrite[i] ^= 1; // duv was changed
		}

		/// <summary>Calculates one iteration of the Jacobi solver and updates the UV by adding DUV.</summary>
		/// <param name="i">The current "level". Actual level is mFineLevel+i.</param>
		protected void CalculateJacobiSolverAndUpdateUV(int i)
		{
			int level = mFineLevel + i;
			int readUV = GetReadTex(mUVwrite[i]);
			int readDUV = GetReadTex(mDUVwrite[i]);
			int writeUV = mUVwrite[i];
			int writeDUV = mDUVwrite[i];

			// activate program and framebuffer object //
			GL.UseProgram(mJacobiProgram);
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, mUpdateUVFBO[i]);

			DrawBuffersEnum[] db = {
				(DrawBuffersEnum)mUVattachpoints[writeUV, i],
				(DrawBuffersEnum)(mDUVattachpoints[writeDUV, i] + 2) // attached to color attachment 2 and 3
			};
			GL.DrawBuffers(2, db);

			// bind textures and set uniforms //
			GL.Uniform2(mJacobiUniLocDxy, mDxy[0, level], -mDxy[1, level]); // - because axis is mirrored in matlab

			BindTextures(new int[] {
				mTensorTexID[0,i],
				mTensorTexID[1,i],
				mTensorTexID[2,i],
				mTensorTexID[3,i],
				mUVTexID[readUV,i],
				mDUVTexID[readDUV,i],
				mPhi3TexID[i],
				//mSilhouetteTexID[i], // [JTompkin] Remove silhouette early culling
			});

			DrawScreenQuad(mMipmapWidth[level], mMipmapHeight[level]);

			GL.ActiveTexture(TextureUnit.Texture0);

			// swap for pingpong
			mUVwrite[i] ^= 1; // uv was changed
			mDUVwrite[i] ^= 1; // duv was changed
		}

		/// <summary>Scales the image by a factor of two from "level" i to i-1. NxN -&gt; 2Nx2N.</summary>
		/// <param name="i">The current "level". Actual level is mFineLevel+i.</param>
		protected void Scale(int i)
		{
			int level = mFineLevel + i - 1; // output resolution level is one larger
			int readUV = GetReadTex(mUVwrite[i]);
			int writeUV = mUVwrite[i - 1];

			// activate program and framebuffer object //
			GL.UseProgram(mScaleProgram);
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, mScaleFBO[i - 1]);

			GL.DrawBuffer(mUVattachpoints[writeUV, i - 1]);
			BindTextures(mUVTexID[readUV, i]);

			DrawScreenQuad(mMipmapWidth[level], mMipmapHeight[level]);

			GL.ActiveTexture(TextureUnit.Texture0);

			mUVwrite[i - 1] ^= 1; // uv was changed
		}

		/// <summary>Scale the result from the iteration step to the actual flow field used for morphing.</summary>
		/// <param name="scaleFactor">The scale factor, with which the input uv flow field is to be multiplied.</param>
		protected void ScaleToResult(int scaleFactor)
		{
			int read = GetReadTex(mUVwrite[mLevelOffset]); // finest level

			// activate program and framebuffer object //
			GL.UseProgram(mScaleToResultProgram);
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, mScaleToResultFBO);
			GL.DrawBuffer(mResultDrawBuffer);

			// bind textures and set uniforms //
			BindTextures(mUVTexID[read, mLevelOffset]);
			GL.Uniform1(mScaleToResultUniLocScale, (float)scaleFactor);

			DrawScreenQuad(mMipmapWidth[0], mMipmapHeight[0]);
		}

		/// <summary>Returns the read index for a ping-pong buffer.</summary>
		/// <param name="val">The write index.</param>
		protected int GetReadTex(int val)
		{
			return val ^ 1;
		}

		/// <summary>Cycles through textures one by one. Useful for debugging.</summary>
		protected int CycleTextures()
		{
			int c = mCycle;
			mCycle++;

			// preprocess
			if (c < 1) return mGrayTexID; else c--;
			if (c < 1) return mGaussTexID; else c--;
			if (c < mCoarseLevel + 1) return mReduceTexID[c]; else c -= (mCoarseLevel + 1);
			if (c < mCoarseLevel) return mReduceHorizontalTexID[c]; else c -= (mCoarseLevel);
			for (int i = 0; i < 5; i++) { if (c < mNumLevels) return mDerivativeTexID[i,c]; else c -= (mNumLevels); }

			// iterate
			for (int i = 0; i < 4; i++) { if (c < mNumLevels) return mTensorTexID[i,c]; else c -= (mNumLevels); }
			if (c < mNumLevels) return mPhi3TexID[c]; else c -= (mNumLevels);
			for (int i = 0; i < 2; i++) { if (c < mNumLevels) return mUVTexID[i,c]; else c -= (mNumLevels); }
			for (int i = 0; i < 2; i++) { if (c < mNumLevels) return mDUVTexID[i,c]; else c -= (mNumLevels); }
			if (c < 3) return mResultUVTexID[c]; else c -= 3;

			// overflow of cycle, need to restart ...

			// undo increment
			mCycle--;

			// subtract number of textures listed above
			mCycle -= (mCycle - c);

			// try again, hopefully not an infinite loop
			return CycleTextures();
		}
		
//-------------------------------------------------------------------------------------------------
		#endregion

		#region Selected OpenGLTools
//-------------------------------------------------------------------------------------------------

		/// <summary>Binds the textures to the currently active FBO.</summary>
		/// <param name="texIDs">The texture IDs of the output texture.</param>
		/// <param name="texAttachments">The texture attachment points.</param>
		private static void BindFBOTextures(int[] texIDs, FramebufferAttachment[] texAttachments)
		{
			// check if graphics card supports this number //
			CheckMaxDrawBuffers(texIDs.Length);

			for (int i = 0; i < texIDs.Length; i++)
			{
				GL.FramebufferTexture2D(FramebufferTarget.Framebuffer, texAttachments[i], TextureTarget.Texture2D, texIDs[i], 0);
			}
		}

		/// <summary>Bind the renderbuffers to the currently active FBO.</summary>
		/// <param name="rbIDs">The renderbuffer IDs.</param>
		/// <param name="rbAttachments">The attachment points for the renderbuffers, like GL_DEPTH_ATTACHMENT_EXT.</param>
		private static void BindRenderbuffers(int[] rbIDs, FramebufferAttachment[] rbAttachments)
		{
			// attach renderbuffers //
			for (int i = 0; i < rbIDs.Length; i++)
			{
				GL.FramebufferRenderbuffer(FramebufferTarget.Framebuffer, rbAttachments[i], RenderbufferTarget.Renderbuffer, rbIDs[i]);
			}
		}

		/// <summary>Binds the given textures on consecutive texture units.</summary>
		/// <param name="texIDs">The texture IDs.</param>
		/// <param name="start">First texture unit.</param>
		private static void BindTextures(int[] texIDs, int start)
		{
			for (int i = texIDs.Length - 1; i >= 0; i--)
			{
				GL.ActiveTexture(TextureUnit.Texture0 + i + start);
				GL.BindTexture(TextureTarget.Texture2D, texIDs[i]);
			}
		}

		/// <summary>Binds the given textures on consecutive texture units, starting at the first.</summary>
		/// <param name="texIDs">The texture IDs.</param>
		private static void BindTextures(int[] texIDs)
		{
			BindTextures(texIDs, 0);
		}

		/// <summary>Binds the given texture to texture unit 0.</summary>
		/// <param name="texIDs">The texture ID.</param>
		private static void BindTextures(int texIDs)
		{
			BindTextures(new int[] { texIDs }, 0);
		}

		/// <summary>Checks the status of the currently bound Framebuffer Object.</summary>
		/// <returns>True if framebuffer complete; else false.</returns>
		private static bool CheckFboStatus()
		{
			switch (GL.CheckFramebufferStatus(FramebufferTarget.Framebuffer))
			{
				case FramebufferErrorCode.FramebufferComplete:
					Trace.WriteLine("FBO: The framebuffer is complete and valid for rendering.");
					return true;

				case FramebufferErrorCode.FramebufferIncompleteAttachment:
					Trace.WriteLine("FBO: One or more attachment points are not framebuffer attachment complete. This could mean there’s no texture attached or the format isn’t renderable. For color textures this means the base format must be RGB or RGBA and for depth textures it must be a DEPTH_COMPONENT format. Other causes of this error are that the width or height is zero or the z-offset is out of range in case of render to volume.");
					break;

				case FramebufferErrorCode.FramebufferIncompleteMissingAttachment:
					Trace.WriteLine("FBO: There are no attachments.");
					break;

				/* case  FramebufferErrorCode.GL_FRAMEBUFFER_INCOMPLETE_DUPLICATE_ATTACHMENT_EXT: 
						 Trace.WriteLine("FBO: An object has been attached to more than one attachment point.");
						 break;
					 */

				case FramebufferErrorCode.FramebufferIncompleteDimensionsExt:
					Trace.WriteLine("FBO: Attachments are of different size. All attachments must have the same width and height.");
					break;

				case FramebufferErrorCode.FramebufferIncompleteFormatsExt:
					Trace.WriteLine("FBO: The color attachments have different format. All color attachments must have the same format.");
					break;

				case FramebufferErrorCode.FramebufferIncompleteDrawBuffer:
					Trace.WriteLine("FBO: An attachment point referenced by GL.DrawBuffers() doesn’t have an attachment.");
					break;

				case FramebufferErrorCode.FramebufferIncompleteReadBuffer:
					Trace.WriteLine("FBO: The attachment point referenced by GL.ReadBuffers() doesn’t have an attachment.");
					break;

				case FramebufferErrorCode.FramebufferUnsupported:
					Trace.WriteLine("FBO: This particular FBO configuration is not supported by the implementation.");
					break;

				default:
					Trace.WriteLine("FBO: Status unknown. (yes, this is really bad.)");
					break;
			}
			return false;
		}

		/// <summary>Checks whether the current graphics card supports enough draw buffers.</summary>
		/// <param name="num">The number of draw buffers one would like to use.</param>
		private static void CheckMaxDrawBuffers(int num)
		{
			int maxDrawBuffers;
			GL.GetInteger(GetPName.MaxDrawBuffers, out maxDrawBuffers);

			if (maxDrawBuffers < num)
				throw new NotSupportedException("Your GPU does not support enough draw buffers.");
		}

		/// <summary>Creates a framebuffer object.</summary>
		/// <param name="texIDs">Textures which gets attached to the GL_COLOR_ATTACHMENTS.</param>
		/// <param name="texAttachments">The texture attachments.</param>
		/// <param name="renderbufferIDs">Renderbuffers which gets attached.</param>
		/// <param name="renderbufferAttachments">The renderbuffer attachments.</param>
		/// <param name="drawbuffers">Drawbuffers.</param>
		/// <returns>ID of the FBO.</returns>
		private static int CreateFBO(int[] texIDs, FramebufferAttachment[] texAttachments, int[] renderbufferIDs, FramebufferAttachment[] renderbufferAttachments, DrawBuffersEnum[] drawbuffers)
		{
			int fboID = 0;

			// generate framebuffer //
			GL.GenFramebuffers(1, out fboID);
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, fboID);

			// attach texture to frambuffer //
			BindFBOTextures(texIDs, texAttachments);

			// attach renderbuffer to framebuffer //
			BindRenderbuffers(renderbufferIDs, renderbufferAttachments);

			// set drawbuffers //
			DrawBuffers(drawbuffers);

			// set parameters //
			GL.Enable(EnableCap.DepthTest);
			GL.DepthFunc(DepthFunction.Lequal);

			GL.Clear(ClearBufferMask.ColorBufferBit);

			CheckFboStatus();

			return fboID;
		}

		/// <summary>Creates a framebuffer object.</summary>
		/// <param name="texIDs">Textures which gets attached to the GL_COLOR_ATTACHMENTs.</param>
		/// <returns>ID of the FBO.</returns>
		private static int CreateFBO(int[] texIDs)
		{
			// texture attachment points
			FramebufferAttachment[] texAtt = Enumerable.Range(0, texIDs.Length).Select(x => (FramebufferAttachment)((int)FramebufferAttachment.ColorAttachment0 + x)).ToArray();

			// no render buffer
			int[] rbIDs = new int[0];
			FramebufferAttachment[] rbAtt = new FramebufferAttachment[0];

			// draw buffers
			DrawBuffersEnum[] db = Enumerable.Range(0, texIDs.Length).Select(x => (DrawBuffersEnum)((int)DrawBuffersEnum.ColorAttachment0 + x)).ToArray();

			return CreateFBO(texIDs, texAtt, rbIDs, rbAtt, db);
		}

		/// <summary>Creates a framebuffer object.</summary>
		/// <param name="texID">Texture which gets attached to the DrawBufferMode.ColorAttachment0.</param>
		/// <returns>ID of the FBO.</returns>
		private static int CreateFBO(int texID)
		{
			return CreateFBO(new int[] { texID } );
		}

		/// <summary>Creates an empty texture as a rendertarget.</summary>
		/// <param name="width">The width of the data.</param>
		/// <param name="height">The height of the data.</param>
		/// <param name="target">The texture target (e.g. GL_TEXTURE2D, GL_TEXTURE_RECTANGLE_ARB).</param>
		/// <param name="internalFormat">The internal format (e.g. GL_RGBA, GL_RGBA16F_ARB, GL_RGB).</param>
		/// <param name="format">The format of the input data (e.g. GL_RGBA, GL_RGB).</param>
		/// <param name="type">The type of the input data (e.g. GL_FLOAT, GL_UNSIGNED_BYTE).</param>
		/// <param name="isMipmap">If set to <c>true</c>, the texture used is a mipmap texture. Not implemented.</param>
		/// <param name="mag_filter">The magnification filter.</param>
		/// <param name="min_filter">The minification filter.</param>
		/// <param name="wrap_s">Wrap mode along x.</param>
		/// <param name="wrap_t">Wrap mode along y.</param>
		private static int CreateFBOTexture(int width, int height, TextureTarget target, PixelInternalFormat internalFormat, PixelFormat format, PixelType type, bool isMipmap, TextureMagFilter mag_filter, TextureMinFilter min_filter, TextureWrapMode wrap_s, TextureWrapMode wrap_t)
		{
			//if (type == PixelType.UnsignedByte)
			//{
			//	return CreateTextureDatauc(NULL, width, height,
			//		target, internalFormat, format, type, false,
			//		mag_filter, min_filter, wrap_s, wrap_t, isMipmap);
			//}

			if (type == PixelType.Float)
			{
				return CreateTextureDataf(IntPtr.Zero, width, height,
					target, internalFormat, format, type, false,
					mag_filter, min_filter, wrap_s, wrap_t, isMipmap);
			}

			throw new NotImplementedException();
		}

		/// <summary>Creates an empty texture as a rendertarget with default settings (RGBA16f, linear, clamp).</summary>
		/// <param name="width">The width of the data.</param>
		/// <param name="height">The height of the data.</param>
		private static int CreateFBOTexture(int width, int height)
		{
			return CreateFBOTexture(width, height, TextureTarget.Texture2D, PixelInternalFormat.Rgba16f, PixelFormat.Rgba, PixelType.Float, false, TextureMagFilter.Linear, TextureMinFilter.Linear, TextureWrapMode.Clamp, TextureWrapMode.Clamp);
		}

		/// <summary>Creates a shader from an embedded resource.</summary>
		/// <param name="name">The name of the embedded resource containing the shader code.</param>
		/// <param name="type">The shader type.</param>
		/// <returns>The shader ID.</returns>
		private static int CreateShader(string name, ShaderType type)
		{
			//// read from file
			//string source = File.ReadAllText(name);
			//int shaderID = GL.CreateShader(type);
			//GL.ShaderSource(shaderID, source);
			//GL.CompileShader(shaderID);

			// read shader from embedded resource
			var shaderID = GL.CreateShader(type);
			using (var shaderStream = Assembly.GetExecutingAssembly().GetManifestResourceStream(typeof(OpticalFlow), name))
			{
				if (shaderStream == null)
					throw new ArgumentException(String.Format("Embedded resource '{0}' does not exist.", name), "name");

				using (var streamReader = new StreamReader(shaderStream))
				{
					GL.ShaderSource(shaderID, streamReader.ReadToEnd());
				}
			}
			GL.CompileShader(shaderID);

			// test for errors
			int status_code;
			GL.GetShader(shaderID, ShaderParameter.CompileStatus, out status_code);
			if (status_code != 1) throw new ApplicationException(GL.GetShaderInfoLog(shaderID));
			Debug.Assert(GL.GetError() == ErrorCode.NoError);

			return shaderID;
		}

		/// <summary>Creates a shader program from various files.</summary>
		/// <param name="vsFilenames">Filenames of the vertex shaders.</param>
		/// <param name="fsFilenames">Filenames of the fragment shaders.</param>
		/// <returns>The program ID.</returns>
		private static int CreateShaderProgram(string[] vsFilenames, string[] fsFilenames)
		{
			// create program //
			int programID = GL.CreateProgram();

			// create and attach shaders //
			foreach (var vsFilename in vsFilenames)
			{
				int shaderID = CreateShader(vsFilename, ShaderType.VertexShader);
				GL.AttachShader(programID, shaderID);
				GL.DeleteShader(shaderID);
			}

			foreach (var fsFilename in fsFilenames)
			{
				int shaderID = CreateShader(fsFilename, ShaderType.FragmentShader);
				GL.AttachShader(programID, shaderID);
				GL.DeleteShader(shaderID);
			}

			// link program //
			GL.LinkProgram(programID);

			// test for errors
			int status_code;
			GL.GetProgram(programID, ProgramParameter.LinkStatus, out status_code);
			if (status_code != 1)
				throw new ApplicationException(GL.GetProgramInfoLog(programID));

			Debug.Assert(GL.GetError() == ErrorCode.NoError);

			return programID;
		}

		/// <summary>Loads a texture from a data array and creates a float texture in OpenGL.</summary>
		/// <param name="data">The data to be assigned to texture.</param>
		/// <param name="width">The width of the data.</param>
		/// <param name="height">The height of the data.</param>
		/// <param name="target">The texture target (e.g. GL_TEXTURE2D, GL_TEXTURE_RECTANGLE_ARB).</param>
		/// <param name="internalFormat">The internal format (e.g. GL_RGBA, GL_RGBA16F_ARB, GL_RGB).</param>
		/// <param name="format">The format of the input data (e.g. GL_RGBA, GL_RGB).</param>
		/// <param name="type">The type of the input data (e.g. GL_FLOAT, GL_UNSIGNED_BYTE).</param>
		/// <param name="isMipmap">If set to <c>true</c>, the texture used is a mipmap texture. Not implemented.</param>
		/// <param name="mag_filter">The magnification filter.</param>
		/// <param name="min_filter">The minification filter.</param>
		/// <param name="wrap_s">Wrap mode along x.</param>
		/// <param name="wrap_t">Wrap mode along y.</param>
		/// <param name="isMipmapRenderTarget">If set to <c>true</c>, the rendertarget is used as a mipmap later on.</param>
		private static int CreateTextureDataf(IntPtr data, int width, int height, TextureTarget target, PixelInternalFormat internalFormat, PixelFormat format, PixelType type, bool isMipmap, TextureMagFilter mag_filter, TextureMinFilter min_filter, TextureWrapMode wrap_s, TextureWrapMode wrap_t, bool isMipmapRenderTarget)
		{
			if (type != PixelType.Float)
				throw new ArgumentException("Wrong type specified. Choose GL_FLOAT.", "type");

			// generate texture id //
			int texID = GL.GenTexture();

			// set parameters //
			GL.BindTexture(TextureTarget.Texture2D, texID);
			GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureWrapS, (int)wrap_s);
			GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureWrapT, (int)wrap_t);

			GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMagFilter, (int)mag_filter);
			GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMinFilter, (int)min_filter);

			// create texture //
			if (isMipmap)
			{
				if (data == IntPtr.Zero)
					throw new ArgumentException("Cannot create mipmap without data.", "data");

				Glu.Build2DMipmap((OpenTK.Graphics.TextureTarget)(int)target, (int)internalFormat, width, height,
					(OpenTK.Graphics.PixelFormat)(int)format, (OpenTK.Graphics.PixelType)(int)type, data);
			}
			else
			{
				GL.TexImage2D(target, 0, internalFormat, width, height, 0, format, type, data);
			}

			// activate mipmap rendertarget //
			if (isMipmapRenderTarget)
			{
				if (isMipmap)
					throw new ArgumentException("Wrong Choice of parameters.");

				if (min_filter == TextureMinFilter.Nearest || min_filter == TextureMinFilter.Linear)
					throw new ArgumentException("Wrong Choice of min_filter, choose GL_X_MIPMAP_X.");

				GL.GenerateMipmap((GenerateMipmapTarget)target);
			}

			return texID;
		}

		/// <summary>Set the draw buffers for the currently active fbo.</summary>
		/// <remarks> Caution: Do not call this, if your currently using standard pipeline.</remarks>
		/// <param name="db">The draw buffers.</param>
		private static void DrawBuffers(DrawBuffersEnum[] db)
		{
			if (db.Length == 0) GL.DrawBuffer(DrawBufferMode.None);
			else GL.DrawBuffers(db.Length, db);
		}

		/// <summary>Draws a screen-filling quad.</summary>
		/// <remarks>
		/// Draws a screen filling quad, to initialize the GPGPU computation.
		/// The currently set viewport is used, if not specified otherwise.
		/// Depth test is disabled, so that the computation is performed for every pixel.
		/// Projection- and ModelView matrices are restored after the computation.
		/// </remarks>
		/// <param name="resx">The viewport resolution in x-direction.</param>
		/// <param name="resy">The viewport resolution in y-direction.</param>
		/// <param name="xmin">The minimum x texture coordinate, which should be used for the lower left corner of the quad.</param>
		/// <param name="ymin">The minimum y texture coordinate, which should be used for the lower left corner of the quad.</param>
		/// <param name="xmax">The maximum x texture coordinate, which should be used for the upper right corner of the quad.</param>
		/// <param name="ymax">The maximum y texture coordinate, which should be used for the upper right corner of the quad.</param>
		private static void DrawScreenQuad(int resx, int resy, float xmin, float ymin, float xmax, float ymax)
		{
			bool depthIsUsed = GL.IsEnabled(EnableCap.DepthTest);
			if (depthIsUsed) GL.Disable(EnableCap.DepthTest);
			GL.PolygonMode(MaterialFace.Front, PolygonMode.Fill);

			// save current projection matrix and init new //
			GL.MatrixMode(MatrixMode.Projection);
			GL.PushMatrix();
			GL.LoadIdentity();

			// save current modelview matrix and init new //
			GL.MatrixMode(MatrixMode.Modelview);
			GL.PushMatrix();
			GL.LoadIdentity();

			// set viewport //
			if (resx >= 0 && resy >= 0)
			{
				GL.PushAttrib(AttribMask.ViewportBit);
				GL.Viewport(0, 0, resx, resy);
			}

			// draw screen filling quad with appropriate texture coordinates //
			GL.Begin(BeginMode.Quads);
			GL.TexCoord2(xmin, ymin); GL.Vertex2(-1, -1);
			GL.TexCoord2(xmax, ymin); GL.Vertex2(1, -1);
			GL.TexCoord2(xmax, ymax); GL.Vertex2(1, 1);
			GL.TexCoord2(xmin, ymax); GL.Vertex2(-1, 1);
			GL.End();

			// reset viewport //
			if (resx >= 0 && resy >= 0)
				GL.PopAttrib();

			// restore old matrices //
			GL.MatrixMode(MatrixMode.Projection);
			GL.PopMatrix();
			GL.MatrixMode(MatrixMode.Modelview);
			GL.PopMatrix();

			// restore depth test //
			if (depthIsUsed) GL.Enable(EnableCap.DepthTest);
		}

		/// <summary>Draws a screen-filling quad.</summary>
		/// <remarks>
		/// Draws a screen filling quad, to initialize the GPGPU computation.
		/// The currently set viewport is used, if not specified otherwise.
		/// Depth test is disabled, so that the computation is performed for every pixel.
		/// Projection- and ModelView matrices are restored after the computation.
		/// </remarks>
		/// <param name="resx">The viewport resolution in x-direction.</param>
		/// <param name="resy">The viewport resolution in y-direction.</param>
		private static void DrawScreenQuad(int resx, int resy)
		{
			DrawScreenQuad(resx, resy, 0.0f, 0.0f, 1.0f, 1.0f);
		}

//-------------------------------------------------------------------------------------------------
		#endregion

//=================================================================================================
		#endregion

		#region Private fields
//=================================================================================================

		// texture specification //
		private TextureTarget       mTarget  = TextureTarget.Texture2D;
		private PixelInternalFormat mIFormat = PixelInternalFormat.Rgba16f;
		private PixelFormat         mFormat  = PixelFormat.Rgba;
		private PixelType           mType    = PixelType.Float;

		private float[,] mDxy;

		// size of the buffers //
		private int mWidth  = 512;
		private int mHeight = 512;
		private int[] mMipmapWidth;
		private int[] mMipmapHeight;

		// clear color //
		private Color4 mClearColor = Color4.Black;

		// number of iterations and alpha beta parameters //
		private int mIterations  = 30; // outer iterations
		private int mJIterations =  1; // Jacobi iterations (formerly Gauss-Seidel)
		private float mAlpha     = 7.6f;
		private float mBeta      = 0.05f;

		// computation levels //
		private int mCoarseLevel = 5;
		private int mFineLevel   = 0;
		private int mNumLevels;
		private int mLevelOffset = 1;

		// silhouette pass //
		private int[] mSilhouetteTexID;

		// rgb2gray pass //
		private int mGrayTexID;
		private int mGrayFBO;
		private int mGrayProgram;

		// Gaussian smoothing //
		private int mGaussTexID;
		private int mGaussFBO;
		private int mGaussProgram;
		private int mGaussUniLocDxy;

		// Reduction pass //
		private int[] mReduceTexID;
		private int[] mReduceHorizontalTexID;
		private int[] mReduceFBO;
		private int mReduceProgram;
		private int mReduceUniLocDxy;

		// derivatives pass //
		private enum Derivative { DX, DY, DXX, DXY, DYY };

		private int[,] mDerivativeTexID; // [derivative,mipmaplevel]
		private int[,] mDerivativeFBO; // all first and all second derivatives can be calculated with MRT in two passes
		private int[] mDerivativeProgram = new int[2];
		private int[] mDerivativeUniLocDxy = new int[2];
		private int[] mDerivativeUniLocDyx = new int[2];

		// tensor pass //
		private int[,] mTensorTexID;
		private int[] mTensorFBO;
		private int mTensorProgram;
		private int mTensorUniLocDxy;

		// phi3 pass //
		private int[] mPhi3TexID;
		private int[] mPhi3FBO;
		private int mPhi3Program;
		private int mPhi3UniLocDxy;

		// Jacobi solver pass //
		private int[] mJacobiFBO;
		private int mJacobiProgram;
		private int mJacobiUniLocDxy;

		// update UV pass //
		private int[] mUpdateUVFBO;
		private int mUpdateUVProgram;

		// scale pass //
		private int[] mScaleFBO;
		private int mScaleProgram;

		// reset UV //
		private int mResetUVFBO;

		// scaleToResult pass //
		private int[] mResultUVTexID = { 0, 0, 0 };
		private int mScaleToResultFBO;
		private int mScaleToResultProgram;
		private int mScaleToResultUniLocScale;
		private DrawBufferMode mResultDrawBuffer;

		// morph pass //
		private int mMorphProgram;
		private int mMorphUniLocValue;

		private int mMorph3Program;
		private int mMorph3UniLocWeights;

		private int mMorphWeightMapProgram;

		private DrawBufferMode[,] mUVattachpoints;
		private DrawBufferMode[,] mDUVattachpoints;

		private int[] mUVwrite;
		private int[] mDUVwrite;

		private int[,] mUVTexID;
		private int[,] mDUVTexID;

		private bool mNeedToUpdateUniforms = true;
		private bool mInitialised = false;
		private int  mCycle;
		
//=================================================================================================
		#endregion
	}
}