﻿// $Id: Texture.cs 538 2011-09-14 19:56:48Z cr333 $
using OpenTK.Graphics.OpenGL;
using System;

namespace VideoLib.OpenGL
{
	/// <summary>Class wrapping a texture handle and providing utility functions.</summary>
	public class Texture
	{
		#region Public properties
//=================================================================================================

		/// <summary>Gets the texture ID.</summary>
		public int ID { get; protected set; }

		/// <summary>Gets the width of the texture.</summary>
		public int Width { get; protected set; }

		/// <summary>Gets the height of the texture.</summary>
		public int Height { get; protected set; }

		/// <summary>Gets or sets the internal pixel format. Default: RGBA16F.</summary>
		public PixelInternalFormat InternalFormat { get; set; }

		/// <summary>Gets or sets the texture magnification filter. Default: linear.</summary>
		public TextureMagFilter MagFilter { get; set; }

		/// <summary>Gets or sets the texture minification filter. Default: linear.</summary>
		public TextureMinFilter MinFilter { get; set; }

		/// <summary>Sets both minification and magnification filters. Write-only.</summary>
		public TextureMagFilter Filter { set { MagFilter = value; MinFilter = (TextureMinFilter)(int)value; } }

		/// <summary>Gets or sets the horizontal texture wrap mode. Default: clamp.</summary>
		public TextureWrapMode WrapS { get; set; }

		/// <summary>Gets or sets the vertical texture wrap mode. Default: clamp.</summary>
		public TextureWrapMode WrapT { get; set; }

		/// <summary>Sets both texture wrap modes. Write-only.</summary>
		public TextureWrapMode Wrap { set { WrapS = value; WrapT = value; } }

		/// <summary>Gets or sets the texture unit this texture is/should be bound to. Default: 0.</summary>
		public TextureUnit Unit { get; set; }

//=================================================================================================
		#endregion // Public properties

		#region Constructors
//=================================================================================================

		/// <summary>Initialises a new texture wrapper, without allocating a texture.</summary>
		/// <remarks>Call <c>Alloc</c> or <c>Create</c> for allocating a texture.</remarks>
		/// <param name="width">The texture width.</param>
		/// <param name="height">The texture height.</param>
		public Texture(int width, int height)
		{
			Width  = width;
			Height = height;

			InternalFormat = PixelInternalFormat.Rgba16f;

			MagFilter = TextureMagFilter.Linear;
			MinFilter = TextureMinFilter.Linear;
			Wrap      = TextureWrapMode.Clamp;
			Unit      = TextureUnit.Texture0;
		}

		/// <summary>Initialises a new texture wrapper from a given texture ID.</summary>
		/// <param name="textureId">The texture ID.</param>
		public Texture(int textureId)
		{
			ID   = textureId;
			Unit = TextureUnit.Texture0;

			// read texture parameters
			int value; // use temporary value, as properties cannot be passed as out parameters
			GL.GetTexLevelParameter(TextureTarget.Texture2D, 0, GetTextureParameter.TextureWidth,          out value); Width          = value;
			GL.GetTexLevelParameter(TextureTarget.Texture2D, 0, GetTextureParameter.TextureHeight,         out value); Height         = value;
			GL.GetTexLevelParameter(TextureTarget.Texture2D, 0, GetTextureParameter.TextureInternalFormat, out value); InternalFormat = (PixelInternalFormat)value;
			GL.GetTexLevelParameter(TextureTarget.Texture2D, 0, GetTextureParameter.TextureMinFilter,      out value); MinFilter      = (TextureMinFilter)value;
			GL.GetTexLevelParameter(TextureTarget.Texture2D, 0, GetTextureParameter.TextureMagFilter,      out value); MagFilter      = (TextureMagFilter)value;
			GL.GetTexLevelParameter(TextureTarget.Texture2D, 0, GetTextureParameter.TextureWrapS,          out value); WrapS          = (TextureWrapMode)value;
			GL.GetTexLevelParameter(TextureTarget.Texture2D, 0, GetTextureParameter.TextureWrapT,          out value); WrapS          = (TextureWrapMode)value;
		}

//=================================================================================================
		#endregion // Constructors

		#region Allocation + deallocation
//=================================================================================================

		/// <summary>Allocates an empty texture according to the configured parameters.</summary>
		/// <returns>Texture ID, for convenience.</returns>
		public int Alloc()
		{
			return Create(IntPtr.Zero, 4, PixelFormat.Rgba, PixelType.Float);
		}

		/// <summary>Applies any changes to the texture filtering and wrapping parameters.</summary>
		public void ApplyParameters()
		{
			GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMagFilter, (int)MagFilter);
			GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMinFilter, (int)MinFilter);
			GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureWrapS, (int)WrapS);
			GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureWrapT, (int)WrapT);
		}

		/// <summary>Creates a texture from the specified image data.</summary>
		/// <param name="data">The image data.</param>
		/// <param name="stride">Number of bytes per row of the image.</param>
		/// <param name="pixelFormat">The pixel format.</param>
		/// <param name="pixelType">Type of the pixel.</param>
		/// <returns>The texture ID, for convenience.</returns>
		public int Create(IntPtr data, int stride, PixelFormat pixelFormat, PixelType pixelType)
		{
			if (ID == 0) ID = GL.GenTexture();
			GL.BindTexture(TextureTarget.Texture2D, ID);
			ApplyParameters();

			// guess unpack alignment from stride
			int unpack = GuessUnpackFromStride(stride);
			GL.PixelStore(PixelStoreParameter.UnpackAlignment, unpack);

			// fill texture with data
			GL.TexImage2D(TextureTarget.Texture2D, 0, InternalFormat, Width, Height, 0, pixelFormat, pixelType, data);

			// cleanup & return
			if(unpack != 4) GL.PixelStore(PixelStoreParameter.UnpackAlignment, 4); // reset to default
			GL.BindTexture(TextureTarget.Texture2D, 0);
			return ID;
		}

		/// <summary>Creates a texture from the specified image data.</summary>
		/// <typeparam name="T">Array data type.</typeparam>
		/// <param name="data">The image data.</param>
		/// <param name="stride">Number of bytes per row of the image.</param>
		/// <param name="pixelFormat">The pixel format.</param>
		/// <param name="pixelType">Type of the pixel.</param>
		/// <returns>The texture ID, for convenience.</returns>
		public int Create<T>(T[] data, int stride, PixelFormat pixelFormat, PixelType pixelType) where T : struct
		{
			if (ID == 0) ID = GL.GenTexture();
			GL.BindTexture(TextureTarget.Texture2D, ID);
			ApplyParameters();

			// guess unpack alignment from stride
			int unpack = GuessUnpackFromStride(stride);
			GL.PixelStore(PixelStoreParameter.UnpackAlignment, unpack);

			// fill texture with data
			GL.TexImage2D(TextureTarget.Texture2D, 0, InternalFormat, Width, Height, 0, pixelFormat, pixelType, data);

			// cleanup & return
			if(unpack != 4) GL.PixelStore(PixelStoreParameter.UnpackAlignment, 4); // reset to default
			GL.BindTexture(TextureTarget.Texture2D, 0);
			return ID;
		}

		/// <summary>Frees the texture.</summary>
		public void Free()
		{
			if (ID != 0)
			{
				GL.DeleteTexture(ID);
				ID = 0;
			}
		}

		/// <summary>Reallocates texture space for a different texture size.</summary>
		/// <param name="width">The new width of the texture.</param>
		/// <param name="height">The new height of the texture.</param>
		/// <returns></returns>
		public int ReAlloc(int width, int height)
		{
			if (width == Width && height == Height && ID != 0)
			{
				return ID;
			}
			else // re-allocate
			{
				Width  = width;
				Height = height;
				return Alloc();
			}
		}

		/// <summary>Updates the texture with the given bitmap data.</summary>
		/// <param name="data">The image data.</param>
		/// <param name="stride">Number of bytes per row of the image.</param>
		/// <param name="pixelFormat">The pixel format.</param>
		/// <param name="pixelType">Type of the pixel.</param>
		public void Update(IntPtr data, int stride, PixelFormat pixelFormat, PixelType pixelType)
		{
			// if the texture is not yet allocated, do so and return
			if (ID == 0)
			{
				Create(data, stride, pixelFormat, pixelType);
			}
			else // update the texture image
			{
				// guess unpack alignment from stride
				int unpack = GuessUnpackFromStride(stride);
				GL.PixelStore(PixelStoreParameter.UnpackAlignment, unpack);

				// update texture
				GL.BindTexture(TextureTarget.Texture2D, ID);
				GL.TexSubImage2D(TextureTarget.Texture2D, 0, 0, 0, Width, Height, pixelFormat, pixelType, data);
				GL.BindTexture(TextureTarget.Texture2D, 0);

				// reset to default
				if (unpack != 4) GL.PixelStore(PixelStoreParameter.UnpackAlignment, 4);
			}
		}

		/// <summary>Updates the texture with the given bitmap data.</summary>
		/// <typeparam name="T">Array data type.</typeparam>
		/// <param name="data">The image data.</param>
		/// <param name="stride">Number of bytes per row of the image.</param>
		/// <param name="pixelFormat">The pixel format.</param>
		/// <param name="pixelType">Type of the pixel.</param>
		public void Update<T>(T[] data, int stride, PixelFormat pixelFormat, PixelType pixelType) where T : struct
		{
			// if the texture is not yet allocated, do so and return
			if (ID == 0)
			{
				Create(data, stride, pixelFormat, pixelType);
			}
			else // update the texture image
			{
				// guess unpack alignment from stride
				int unpack = GuessUnpackFromStride(stride);
				GL.PixelStore(PixelStoreParameter.UnpackAlignment, unpack);

				// update texture
				GL.BindTexture(TextureTarget.Texture2D, ID);
				GL.TexSubImage2D(TextureTarget.Texture2D, 0, 0, 0, Width, Height, pixelFormat, pixelType, data);
				GL.BindTexture(TextureTarget.Texture2D, 0);

				// reset to default
				if (unpack != 4) GL.PixelStore(PixelStoreParameter.UnpackAlignment, 4);
			}
		}

//=================================================================================================
		#endregion // Allocation + deallocation

		#region Utility functions
//=================================================================================================

		/// <summary>Binds this texture to the current texture unit.</summary>
		public void Bind()
		{
			GL.ActiveTexture(Unit);
			GL.BindTexture(TextureTarget.Texture2D, ID);
		}

		/// <summary>Binds this texture to the given texture unit.</summary>
		/// <param name="unit">The texture unit.</param>
		public void Bind(TextureUnit unit)
		{
			Unit = unit;
			GL.ActiveTexture(unit);
			GL.BindTexture(TextureTarget.Texture2D, ID);
		}

		/// <summary>Unbinds this texture.</summary>
		public void Unbind()
		{
			GL.ActiveTexture(Unit);
			GL.BindTexture(TextureTarget.Texture2D, 0);
		}

		/// <summary>Performs an implicit conversion from <see cref="VideoLib.OpenGL.Texture"/> to <see cref="System.Int32"/> texture ID.</summary>
		/// <param name="texture">The wrapped texture.</param>
		/// <returns>The texture ID.</returns>
		public static implicit operator int(Texture texture)
		{
			return texture.ID;
		}

		/// <summary>Returns a <see cref="System.String"/> that represents this instance.</summary>
		/// <returns>A <see cref="System.String"/> that represents this instance.</returns>
		public override string ToString()
		{
			return String.Format("Texture #{0} ({1}x{2} {3})", ID, Width, Height, InternalFormat);
		}

//=================================================================================================
		#endregion // Utility functions

		#region BitmapFrame-specific convenience functions
//=================================================================================================

		/// <summary>Initialises a new instance of the <see cref="Texture"/> class based on the size of the given bitmap frame.</summary>
		/// <remarks>Note that the texture is not allocated.</remarks>
		/// <param name="bitmapFrame">The bitmap frame.</param>
		public Texture(VideoLib.Frames.BitmapFrame bitmapFrame) : this(bitmapFrame.Width, bitmapFrame.Height)
		{
		}

		/// <summary>Creates a texture from the given bitmap frame.</summary>
		/// <param name="bitmapFrame">The bitmap frame.</param>
		/// <returns>The texture ID, for convenience.</returns>
		public int Create(VideoLib.Frames.BitmapFrame bitmapFrame)
		{
			Width  = bitmapFrame.Width;
			Height = bitmapFrame.Height;
			return Create(bitmapFrame.Data, bitmapFrame.Stride, bitmapFrame.Format.ToOpenGLPixelFormat(), bitmapFrame.Format.ToOpenGLPixelType());
		}

		/// <summary>Reads the contents of the texture into a bitmap frame of the given format.</summary>
		/// <param name="format">The desired pixel format.</param>
		public VideoLib.Frames.BitmapFrame ReadFrame(VideoLib.Frames.PixelFormat format)
		{
			Bind();

			// create output frame
			VideoLib.Frames.BitmapFrame frame = new VideoLib.Frames.BitmapFrame(Width, Height, format);

			// read out texture
			GL.PixelStore(PixelStoreParameter.PackAlignment, 4);
			GL.GetTexImage(TextureTarget.Texture2D, 0, format.ToOpenGLPixelFormat(), format.ToOpenGLPixelType(), frame.Data);

			Unbind();

			return frame;
		}

		/// <summary>Updates this texture from the given bitmap frame. Reallocates texture if necessary.</summary>
		/// <param name="bitmapFrame">The bitmap frame.</param>
		public void Update(VideoLib.Frames.BitmapFrame bitmapFrame)
		{
			// if the bitmap has the correct size, just update the texture in place
			if (Width == bitmapFrame.Width && Height == bitmapFrame.Height)
			{
				// update texture
				Update(bitmapFrame.Data, bitmapFrame.Stride, bitmapFrame.Format.ToOpenGLPixelFormat(), bitmapFrame.Format.ToOpenGLPixelType());
			}
			else // need to create new image
			{
				Width  = bitmapFrame.Width;
				Height = bitmapFrame.Height;
				Create(bitmapFrame.Data, bitmapFrame.Stride, bitmapFrame.Format.ToOpenGLPixelFormat(), bitmapFrame.Format.ToOpenGLPixelType());
			}
		}

//=================================================================================================
		#endregion // BitmapFrame-specific convenience functions

		#region Internal utilities
//=================================================================================================

		private static int GuessUnpackFromStride(int stride)
		{
			if (stride % 8 == 0) return 8;
			if (stride % 4 == 0) return 4;
			if (stride % 2 == 0) return 2;
			return 1;
		}

		private static int GuessUnpackFromFormat(PixelFormat pixelFormat, PixelType pixelType)
		{
			if (pixelFormat == PixelFormat.Rgb) return 4;
			if (pixelFormat == PixelFormat.Bgr) return 4;
			if (pixelType == PixelType.UnsignedShort) return 2;
			if (pixelType == PixelType.UnsignedByte)  return 1;
			return 4;
		}

//=================================================================================================
		#endregion // Internal utilities
	}
}