﻿// $Id: Scheduler.cs 549 2012-01-17 18:12:19Z cr333 $
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;

namespace VideoLib.Scheduling
{
	///<summary>Scheduler statuses.</summary>
	public enum SchedulerStatus
	{
		/// <summary>The scheduler is instantiated but has not been started yet.</summary>
		Offline,

		/// <summary>The scheduler is starting up.</summary>
		Starting,

		/// <summary>The scheduler is running.</summary>
		Running,

		/// <summary>The scheduler is stopping.</summary>
		Stopping,

		/// <summary>The scheduler has stopped.</summary>
		Stopped
	}

	/// <summary>Base class of any scheduler.</summary>
	public abstract class Scheduler
	{
		/// <summary>Current status of the scheduler.</summary>
		public SchedulerStatus Status { get; protected set; }

		/// <summary>Underlying topology of nodes.</summary>
		protected Topology Topology { get; set; }
		
		/// <summary>List of worker threads.</summary>
		protected List<Thread> workers = new List<Thread>();

		/// <summary>Profiler used internally.</summary>
		public IProfiler Profiler { get; protected set; }

		/// <summary>Whether profiling is enabled.</summary>
		public bool Profiling { get; set; }

		/// <summary>Constructs an abstract Scheduler.</summary>
		/// <param name="topology">Topology of nodes to use.</param>
		public Scheduler(Topology topology)
		{
			Status = SchedulerStatus.Offline;
			Topology = topology;
			Profiling = false;
		}

		/// <summary>Starts a new worker thread.</summary>
		/// <param name="workerID">The worker thread ID.</param>
		/// <returns>Thread handle of the created worker thread.</returns>
		protected virtual Thread StartWorker(int workerID)
		{
			// start a new scheduler thread
			Thread worker = new Thread(RunWorker) { Name = String.Format("Worker{0}", workerID) };
			worker.Start(workerID);
			workers.Add(worker);
			return worker;
		}

		/// <summary>Starts a number of worker threads.</summary>
		/// <param name="numWorkers">Number of worker threads to start.</param>
		public virtual void StartWorkers(int numWorkers)
		{
			for (int i = 0; i <= numWorkers - 1; i++)
				StartWorker(i);
		}

		/// <summary>Sets up the scheduler, preparing it for the work ahead.</summary>
		public virtual void Setup()
		{
			if (Profiling)
				Profiler = new BasicProfiler();

			Topology.CheckConnections();
		}

		/// <summary>Main function of a worker thread. This is where the scheduling happens.</summary>
		/// <param name="data">The thread's worker ID, if <see cref="Scheduler.StartWorker"/> is not not overidden.</param>
		protected abstract void RunWorker(object data);

		/// <summary>Stops the scheduler. Experimental.</summary>
		public virtual void Stop()
		{
			// ignore multiple calls to Stop()
			if (Status == SchedulerStatus.Stopping || Status == SchedulerStatus.Stopped) return;

			Status = SchedulerStatus.Stopping;

			// wait till all worker threads stopped
			foreach (var thread in workers)
			{
				// don't wait for the current thread (that is if Stop is called by a worker thread)
				if (!thread.Equals(Thread.CurrentThread))
					thread.Join();
			}

			Debug.WriteLine("Scheduler.Stop: Disposing nodes.");
			foreach (INode node in Topology.Nodes)
			{
				node.Stop();
				node.Dispose();
			}

			Debug.WriteLine("Scheduler.Stop: Empty buffer.");
			foreach (InputPin pin in Topology.IncomingEdges.Keys)
				pin.ClearBuffer();
			
			Topology.Clear();

			Debug.WriteLine("Scheduler.Stop: Stopped scheduler.");
			Status = SchedulerStatus.Stopped;

			if (Profiling)
			{
				PrintProfilingData();
				LogProfilingData();
			}
		}

		/// <summary>
		/// Waits for all worker threads to finish and the scheduler to stop cleanly.
		/// </summary>
		public void Join()
		{
			// wait for all nodes in the topology to finish
			Topology.Join();

			// ignore if Stop() has been called
			if (Status == SchedulerStatus.Stopping || Status == SchedulerStatus.Stopped) return;
			Status = SchedulerStatus.Stopping;

			// wait for all worker threads to exit
			foreach (var thread in workers)
				thread.Join();

			Debug.WriteLine("Scheduler.Join: Disposing nodes.");
			foreach (INode node in Topology.Nodes)
			{
				node.Stop();
				node.Dispose();
			}

			Debug.WriteLine("Scheduler.Join: Finished scheduler.");
			Status = SchedulerStatus.Stopped;

			if (Profiling)
			{
				PrintProfilingData();
				LogProfilingData();
			}
		}

		/// <summary>Prints a summary of the profiling data.</summary>
		protected void PrintProfilingData()
		{
			if (Profiler == null) return;

			// prepare profiling targets
			Profiler.PrepareResults();
			var targets = Profiler.GetProfilingTargets();

			// find the bottlenecks
			var results   = targets.Select(x => Profiler.GetResult(x));
			var intervals = results.Where(x => x.Intervals.Count() > 0);
			var higestRuns  = (intervals.Count() <= 0 ? 0   : intervals.Max(x => x.Intervals.Count()));
			var lowestFPS   = (intervals.Count() <= 0 ? 0.0 : intervals.Min(x => 1 / x.Intervals.Average()));
			var higestMean  = (results.Count() <= 0 ? 0.0 : results.Max(x => x.RunTimeMean));
			var higestStdev = (results.Count() <= 0 ? 0.0 : results.Max(x => x.RunTimeStDev));
			var averageFPS  = (intervals.Count() <= 0 ? 0.0 : intervals.Sum(x => x.Intervals.Count()) / intervals.Sum(x => x.Intervals.Sum()));
			
			// build table of profiling results, ordered by node name
			StringBuilder result = new StringBuilder(Environment.NewLine);
			foreach (Object node in targets.OrderBy(x => x.ToString()))
			{
				ProfilingData d = Profiler.GetResult(node);
				double fps = (d.Intervals.Count() > 0 ? 1 / d.Intervals.Average() : 0);
				result.AppendFormat("{0,-35}: {1,4:D}x{2} | {3,6:F2}{4} ± {5,6:F2} ms{6} | {7,5:F2} fps{8}\n",
					node.ToString(),
					d.Intervals.Count(),   (d.Intervals.Count() == higestRuns  ? "*" : " "),
					d.RunTimeMean  * 1000, (d.RunTimeMean       == higestMean  ? "*" : " "),
					d.RunTimeStDev * 1000, (d.RunTimeStDev      == higestStdev ? "*" : " "),
					fps, (fps == lowestFPS ? "*" : " "));
			}
			result.AppendLine("-------------------------------------------------------------------------------");
			result.AppendFormat("                                                                    {0,6:F2} fps\n", averageFPS);

			Console.WriteLine(result.ToString());
		}

		/// <summary>Saves aggregated profiling data as a CSV file.</summary>
		protected void LogProfilingData()
		{
			if (Profiler == null) return;

			// essentially unique filename for CSV file
			var filename = Path.Combine("profiling", "profile-log_" + DateTime.Now.ToString("yyyy-MM-dd_HH-mm-ss") + ".csv");
			if (!Directory.Exists("profiling")) Directory.CreateDirectory("profiling");

			using (var writer = new StreamWriter(filename))
			{
				writer.WriteLine("Node,Runs,AvgRunTime,StDevRunTime,AvgFPS");

				// prepare profiling targets
				Profiler.PrepareResults();
				var targets = Profiler.GetProfilingTargets();
				
				// output table of profiling results, ordered by node name
				foreach (Object node in targets.OrderBy(x => x.ToString()))
				{
					ProfilingData d = Profiler.GetResult(node);
					double fps = (d.Intervals.Count() > 0 ? 1 / d.Intervals.Average() : 0);
					writer.WriteLine("\"{0}\",{1},{2:F3},{3:F3},{4:F3}", node.ToString(), d.Intervals.Count(), d.RunTimeMean  * 1000, d.RunTimeStDev * 1000, fps);
				}
			}
		}
	}
}
